package rbas.ranksearch.sem;

import edu.cmu.tetrad.util.MatrixUtils;
import edu.cmu.tetrad.graph.*;
import edu.cmu.tetrad.data.RectangularDataSet;

import java.io.Serializable;
import java.util.Iterator;
import java.util.List;
import java.util.ArrayList;
import java.util.Random;

import rbas.util.RBASProbUtils;

/**
 * Implements a variational approximation for a mixture of Gaussian SEMs. This is very closed related to
 * Ghaharamani and Beal's variational mixture of factor analyzers. This also implements special Structural EM-like
 * evaluations for the search operators of the RankSearch class.
 *
 * For more information:
 *
 *  - Silva, R. and Scheines, R. (2006) "Bayesian learning of measurement and structural models". ICML
 *  - Silva, R. (2005) Automatic Discovery of Latent Variable Models, PhD Thesis, Machine Learning Department, CMU
 *
 * @author Ricardo Silva, Fall 2004 - Fall 2006
 * @version $Revision: 1.1 $ $Date: 2006/09/09 19:42:51 $
 */


public class VariationalSemMixture implements Serializable {

    RectangularDataSet dataSet;
    Graph   graph, originalGraph;
    int     numLatents, numMainLatents, numConfounders, numVars, sampleSize;
    int     varsIdx[];

    double  rawData[][];
    Node    latents[], indicators[];
    Edge    confounders[];                                                            
    int     xxParents[][], xyParents[][], yyParents[][], yxParents[][];
    boolean isConfounder[];

    //Model hyperparameters and priors
    double alphaMStar[];
    double betaMuInvCovPrior[][][][], betaMuCovPrior[][][][], betaMuMeanPrior[][][], betaMuInvCovPriorParam;
    double xInvCovPrior[][][], xMeanPrior[][];
    double lambdaInvCovPrior[][][][], lambdaCovPrior[][][][], lambdaMeanPrior[][][], lambdaInvCovPriorParam[];
    boolean singleInvCovParam;
    //Variational distribution parameters
    double qAlphaM[];                                 //parameters of q(Pi)
    double qS[][];                                    //parameters of q(S)
    double qXMean[][][], qXCov[][][][];               //marginal parameters of q(Xi, Zi | S)
    double qMainXCov[][][][];
    double qEX[][][], qEXX[][][][];                   //marginal implied moments of q(Xi, Zi | S)
    double qBetaMuMean[][][], qBetaMuCov[][][][];     //parameters of q(Beta, Mu)
    double qEBM[][][], qEBMBM[][][][];                //implied moments of q(Beta, Mu)
    double upsilon[][];                               //latent variance
    double qLambdaMean[][][], qLambdaCov[][][][];     //parameters for q(lambda)
    double qELambda[][][], qELambdaLambda[][][][];    //implied moments for q(upsilon)
    double qInvPsiA[], qInvPsiB[];
    //double invPsiA, invPsiB;
    double qEInvPsi[];                                //error term variances for each variables (the same for all mixtures)

    double likelihoodKernels[][][];
    double sumS[];

    double score = -Double.MAX_VALUE;

    int numMixtures;

    protected static int    MAX_ITER = 50;
    protected static double MIN_TOLERANCE = 1.e-5;

    protected boolean converging = true;

    List freeVariables;

    protected boolean usePriors = true, updatePriors = true, useBetaPriors = true;
    protected int numParameters;
    protected List givenPriors = null;

    public VariationalSemMixture(RectangularDataSet dataSet, Graph graph, List freeVariables, int numMixtures) {
        this.dataSet = dataSet;
        this.graph = new EdgeListGraph(graph);
        this.originalGraph = new EdgeListGraph(graph);
        this.freeVariables = freeVariables;
        this.numMixtures = numMixtures;
        initialize();
    }

    private void initialize() {
        //Graph
        for (Iterator it = this.freeVariables.iterator(); it.hasNext();) {
            Node newNode = new GraphNode(it.next().toString());
            this.graph.addNode(newNode);
            for (Iterator it2 = this.graph.getNodes().iterator(); it2.hasNext();) {
                Node nextNode = (Node) it2.next();
                if (nextNode != newNode && nextNode.getNodeType() == NodeType.MEASURED)
                    this.graph.addDirectedEdge(nextNode, newNode);
            }
        }
        List measures = new ArrayList();
        List latentsList = new ArrayList();
        for (Iterator it = this.graph.getNodes().iterator(); it.hasNext();) {
            Node next = (Node) it.next();
            if (next.getNodeType() == NodeType.MEASURED)
                measures.add(next);
            else if (next.getNodeType() == NodeType.LATENT)
                latentsList.add(next);
        }
        List confoundersList = new ArrayList();
        for (Iterator it = this.graph.getEdges().iterator(); it.hasNext();) {
            Edge next = (Edge) it.next();
            if (next.getEndpoint1() == Endpoint.ARROW && next.getEndpoint2() == Endpoint.ARROW)
                confoundersList.add(next);
        }
        this.numConfounders = confoundersList.size();
        this.confounders = new Edge[this.numConfounders];
        for (int i = 0; i < this.numConfounders; i++)
             this.confounders[i] = (Edge) confoundersList.get(i);

        this.numVars = measures.size();
        this.numLatents = latentsList.size() + confoundersList.size();
        this.numMainLatents = latentsList.size();
        this.isConfounder = new boolean[this.numLatents];
        this.latents = new Node[this.numLatents];
        for (int i = 0; i < this.numLatents; i++) {
            if (i < this.numMainLatents) {
                this.latents[i] = (Node) latentsList.get(i);
                this.isConfounder[i] = false;
            }
            else {
                this.latents[i] = new GraphNode(confoundersList.get(i - this.numMainLatents).toString());
                this.isConfounder[i] = true;
            }
        }

        this.numParameters = this.numVars + (this.numMixtures - 1);
        this.indicators = new Node[this.numVars];
        this.varsIdx = new int[this.numVars];
        this.yxParents = new int[this.numVars][];
        List varNamesList = this.dataSet.getVariableNames();
        String variableNames[] = new String[varNamesList.size()];
        for (int i = 0; i < variableNames.length; i++)
            variableNames[i] = (String) varNamesList.get(i);
        for (int i = 0; i < this.numVars; i++) {
            this.indicators[i] = (Node) measures.get(i);
            for (int v = 0; v < this.dataSet.getNumColumns(); v++)
                if (this.indicators[i].toString().equals(variableNames[v])) {
                    this.varsIdx[i] = v;
                    break;
                }
            List xP = new ArrayList();
            for (Iterator it = this.graph.getParents(this.indicators[i]).iterator(); it.hasNext();) {
                Node nextParent = (Node) it.next();
                if (nextParent.getNodeType() == NodeType.LATENT)
                    xP.add(nextParent);
            }
            for (int c = 0; c < this.confounders.length; c++) {
                if (this.confounders[c].getNode1() == this.indicators[i] ||
                        this.confounders[c].getNode2() == this.indicators[i])
                    xP.add(new Integer(c));
            }

            this.yxParents[i] = new int[xP.size()];
            this.numParameters += xP.size() * this.numMixtures;
            int countPx = 0;
            for (Iterator it = xP.iterator(); it.hasNext();) {
                Object next = it.next();
                if (next instanceof Integer)
                    this.yxParents[i][countPx++] = this.numMainLatents + ((Integer) next).intValue();
                else
                    for (int j = 0; j < this.latents.length; j++)
                        if (this.latents[j].toString().equals(next.toString())) {
                            this.yxParents[i][countPx++] = j;
                            break;
                        }
            }
        }
        this.yyParents = new int[this.numVars][];
        for (int i = 0; i < this.numVars; i++) {
            List yP = new ArrayList();
            for (Iterator it = this.graph.getParents(this.indicators[i]).iterator(); it.hasNext();) {
                Node nextParent = (Node) it.next();
                if (nextParent.getNodeType() == NodeType.MEASURED)
                    yP.add(nextParent);
            }
            this.yyParents[i] = new int[yP.size() + 1];
            this.numParameters += (yP.size() + 1) * this.numMixtures;
            int countP = 0;
            for (Iterator it = yP.iterator(); it.hasNext();) {
                Node next = (Node) it.next();
                for (int j = 0; j < this.indicators.length; j++)
                    if (this.indicators[j].toString().equals(next.toString())) {
                        this.yyParents[i][countP++] = j;
                        break;
                    }
            }
            this.yyParents[i][yP.size()] = this.numVars; //this will correspond to the intercept (mu) term
        }

        this.xxParents = new int[this.numLatents][];
        this.xyParents = new int[this.numLatents][];
        for (int i = 0; i < this.numLatents; i++) {
            if (i >= this.numMainLatents) {
                this.xxParents[i] = new int[0];
                this.xyParents[i] = new int[0];
                continue;
            }
            List xP = this.graph.getParents(this.latents[i]);
            List xPX = new ArrayList();
            List xPY = new ArrayList();
            for (Iterator it = xP.iterator(); it.hasNext();) {
                Node next = (Node) it.next();
                if (next.getNodeType() == NodeType.LATENT)
                    xPX.add(next);
                else
                    xPY.add(next);
            }
            this.xyParents[i] = new int[xPY.size()];
            this.numParameters += xP.size() * this.numMixtures;
            int countPy = 0;
            for (Iterator it = xPY.iterator(); it.hasNext();) {
                Node next = (Node) it.next();
                for (int j = 0; j < this.indicators.length; j++)
                    if (this.indicators[j].toString().equals(next.toString())) {
                        this.xyParents[i][countPy++] = j;
                        break;
                    }
            }
            this.xxParents[i] = new int[xPX.size()];
            this.numParameters += xPX.size() * this.numMixtures;
            int countP = 0;
            for (Iterator it = xPX.iterator(); it.hasNext();) {
                Node next = (Node) it.next();
                for (int j = 0; j < this.latents.length; j++)
                    if (this.latents[j].toString().equals(next.toString())) {
                        this.xxParents[i][countP++] = j;
                        break;
                    }
            }
        }

        //Data
        this.sampleSize = this.dataSet.getNumRows();
        this.rawData = new double[this.sampleSize][this.numVars + 1];
        for (int i = 0; i < this.numVars; i++) {
            int vPos = -1;
            for (int p = 0; p < variableNames.length; p++)
                if (this.indicators[i].toString().equals(variableNames[p])) {
                    vPos = p;
                    break;
                }
            double raw[] = this.dataSet.getDoubleMatrix().viewColumn(vPos).toArray();
            for (int j = 0; j < this.sampleSize; j++)
                this.rawData[j][i] = raw[j];
        }
        for (int j = 0; j < this.sampleSize; j++)
            this.rawData[j][this.numVars] = 1.;

        //Parameters for Pi
        this.qAlphaM = new double[this.numMixtures];
        this.alphaMStar = new double[this.numMixtures];
        //Parameters for S
        this.qS = new double[this.numMixtures][this.sampleSize];

        //Parameters for X, Z
        this.betaMuMeanPrior = new double[this.numMixtures][this.numLatents][];
        this.betaMuCovPrior = new double[this.numMixtures][this.numLatents][][];
        this.betaMuInvCovPrior = new double[this.numMixtures][this.numLatents][][];
        this.qBetaMuMean = new double[this.numMixtures][this.numLatents][];
        this.qBetaMuCov = new double[this.numMixtures][this.numLatents][][];
        this.qEBM = new double[this.numMixtures][this.numLatents][];
        this.qEBMBM = new double[this.numMixtures][this.numLatents][][];
        for (int d = 0; d < this.numMixtures; d++)
            for (int x = 0; x < this.numLatents; x++) {
                int numParents = this.xxParents[x].length + this.xyParents[x].length;
                this.betaMuMeanPrior[d][x] = new double[numParents];
                this.qBetaMuMean[d][x] = new double[numParents];
                this.betaMuCovPrior[d][x] = new double[numParents][numParents];
                this.betaMuInvCovPrior[d][x] = new double[numParents][numParents];
                this.qBetaMuCov[d][x] = new double[numParents][numParents];
                this.qEBM[d][x] = new double[numParents];
                this.qEBMBM[d][x] = new double[numParents][numParents];
            }

        this.upsilon = new double[this.numMixtures][this.numLatents];
        this.xInvCovPrior = new double[this.numMixtures][this.numLatents][this.numLatents];
        this.xMeanPrior = new double[this.numMixtures][this.numLatents];

        this.qXMean = new double[this.numMixtures][this.sampleSize][this.numLatents];
        this.qXCov = new double[this.numMixtures][this.sampleSize][this.numLatents][this.numLatents];
        this.qMainXCov = new double[this.numMixtures][this.sampleSize][this.numMainLatents][this.numMainLatents];
        this.qEX = new double[this.numMixtures][this.sampleSize][this.numLatents];
        this.qEXX = new double[this.numMixtures][this.sampleSize][this.numLatents][this.numLatents];

        //Parameters for Y

        this.lambdaInvCovPriorParam = new double[this.numVars];
        this.lambdaMeanPrior = new double[this.numMixtures][this.numVars][];
        this.lambdaCovPrior = new double[this.numMixtures][this.numVars][][];
        this.lambdaInvCovPrior = new double[this.numMixtures][this.numVars][][];
        this.qLambdaMean = new double[this.numMixtures][this.numVars][];
        this.qLambdaCov = new double[this.numMixtures][this.numVars][][];
        this.qELambda = new double[this.numMixtures][this.numVars][];
        this.qELambdaLambda = new double[this.numMixtures][this.numVars][][];
        for (int d = 0; d < this.numMixtures; d++)
            for (int y = 0; y < this.numVars; y++) {
                int numParents = this.yyParents[y].length + this.yxParents[y].length;
                this.lambdaMeanPrior[d][y] = new double[numParents];
                this.lambdaInvCovPrior[d][y] = new double[numParents][numParents];
                this.lambdaCovPrior[d][y] = new double[numParents][numParents];
                this.qLambdaMean[d][y] = new double[numParents];
                this.qLambdaCov[d][y] = new double[numParents][numParents];
                this.qELambda[d][y] = new double[numParents];
                this.qELambdaLambda[d][y] = new double[numParents][numParents];
            }
        this.qInvPsiA = new double[this.numVars]; this.qInvPsiB = new double[this.numVars];
        this.qEInvPsi = new double[this.numVars];

        //Other
        this.likelihoodKernels = new double[this.numMixtures][this.sampleSize][this.numVars];
        this.sumS = new double[this.numMixtures];
        this.singleInvCovParam = false;
        createStructuralEmVariables();
    }

    public void setSingleInvCovParam(boolean invCov) {
        this.singleInvCovParam = invCov;
    }

    public void setUsePriors(boolean usePriors) {
        this.usePriors = usePriors;
    }

    public void estimate() {
        this.converging = true;
        initializeParameters();
        double f, newF = fEval();
        int iter = 0;
        //System.out.println(" ------ F0 = " + newF + " " + this.numMixtures);
        do {
            do {
                f = newF;
                //updatePsiHyperParameters();
                //newF = fEval(); System.out.println(" ------ F (Psi)     = " + newF);
                //updateLambdaHyperParameters(); updateBetaMuHyperParameters(); //updatePsiHyperParameters();
                //newF = fEval(); System.out.println(" ------ F (Upsilon) = " + newF);
                updateQX();
                  //newF = fEval(); System.out.println(" ------ F (X)       = " + newF);
                updateQS();
                  //newF = fEval(); System.out.println(" ------ F (S)       = " + newF);
                updateQPi();
                  //newF = fEval(); System.out.println(" ------ F (Pi)      = " + newF);
                updateQPsi(); updateQLambda(); updateQPsi(); updateLambdaHyperParameters();
                  //newF = fEval(); System.out.println(" ------ F (Lambda)  = " + newF);
                updateQUpsilon(); updateQBetaMu(); updateQUpsilon(); updateBetaMuHyperParameters();
                  //newF = fEval(); System.out.println(" ------ F (Beta)    = " + newF);
                  //newF = fEval(); System.out.println(" ------ F (BetaH)   = " + newF);
                  //newF = fEval(); System.out.println(" ------ F (X)       = " + newF);
                  //updateQS();
                newF = fEval();
                //System.out.println(" ------ F = " + newF + " " + this.numMixtures);
                iter++;
            } while (this.converging && newF - f > MIN_TOLERANCE && iter < MAX_ITER);
            //updateLambdaHyperParameters(); updateBetaMuHyperParameters(); //updatePsiHyperParameters();
            //newF = fEval();
        } while (this.converging && newF - f > MIN_TOLERANCE && iter < MAX_ITER);
        if (!this.converging || newF < f) {
            //this.score = -Double.MAX_VALUE;
            //fEvalDebug();
            //System.exit(0);
            System.out.println("BOOM!");
        }
        else
            this.score = newF;
    }

    private void initializeParameters() {
        Random r = new Random();
        int mult;
        if (this.usePriors)
            mult = 1;
        else
            mult = 0;
        //Pi priors
        for (int s = 0; s < this.numMixtures; s++)
            this.alphaMStar[s] = mult * 10. / this.numMixtures;
        //Beta/Mu/Upsilon priors
        this.betaMuInvCovPriorParam = r.nextDouble();
        //this.betaMuInvCovPriorParam = 10.;
        for (int d = 0; d < this.numMixtures; d++)
            for (int l = 0; l < this.numLatents; l++) {
                int numParents = this.xxParents[l].length + this.xyParents[l].length;
                for (int v1 = 0; v1 < numParents; v1++) {
                    this.betaMuMeanPrior[d][l][v1] = 0.; //This *has* to be zero.
                    for (int v2 = 0; v2 < numParents; v2++) {
                        if (v1 == v2) {
                            this.betaMuInvCovPrior[d][l][v1][v2] = mult * this.betaMuInvCovPriorParam;
                            this.betaMuCovPrior[d][l][v1][v2] = mult * 1. / this.betaMuInvCovPriorParam;
                        }
                        else {
                            this.betaMuCovPrior[d][l][v1][v2] = 0.;
                            this.betaMuInvCovPrior[d][l][v1][v2] = 0.;
                        }
                    }
                }
            }
        //Lambda priors
        if (this.updatePriors) {
            if (this.singleInvCovParam) {
                this.lambdaInvCovPriorParam[0] = r.nextDouble() + 0.2;
                for (int y = 1; y < this.numVars; y++)
                    this.lambdaInvCovPriorParam[y] = this.lambdaInvCovPriorParam[0];
            }
            else for (int y = 0; y < this.numVars; y++)
                this.lambdaInvCovPriorParam[y] = r.nextDouble() + 0.2;
        } else {
            for (int i = 0; i < this.givenPriors.size(); i += 2) {
                String name = this.givenPriors.get(i).toString();
                double value = ((Double) this.givenPriors.get(i + 1)).doubleValue();
                for (int j = 0; j < this.indicators.length; j++)
                    if (this.indicators[j].getName().equals(name)) {
                        this.lambdaInvCovPriorParam[j] = value;
                        break;
                    }
            }
        }
        for (int d = 0; d < this.numMixtures; d++)
            for (int y = 0; y < this.numVars; y++) {
                int numParents = this.yyParents[y].length + this.yxParents[y].length;
                for (int v1 = 0; v1 < numParents; v1++) {
                    this.lambdaMeanPrior[d][y][v1] = 0.; //This *has* to be zero.
                    for (int v2 = 0; v2 < numParents; v2++) {
                        if (v1 == v2) {
                            this.lambdaInvCovPrior[d][y][v1][v2] = mult * this.lambdaInvCovPriorParam[y];
                            this.lambdaCovPrior[d][y][v1][v2] = mult * 1. / this.lambdaInvCovPriorParam[y];
                        }
                        else {
                            this.lambdaInvCovPrior[d][y][v1][v2] = 0.;
                            this.lambdaCovPrior[d][y][v1][v2] = 0.;
                        }

                    }
                }
                this.lambdaCovPrior[d][y] = MatrixUtils.inverse(this.lambdaInvCovPrior[d][y]);
            }
        for (int s = 0; s < this.numMixtures; s++)
            //this.qAlphaM[s] = mult * r.nextDouble() * 10.;
            this.qAlphaM[s] = mult * 10.;

        //Q distribution parameters
        for (int i = 0; i < this.sampleSize; i++) {
            double qSum = .0;
            for (int s = 0; s < this.numMixtures; s++) {
                //this.qS[s][i] = r.nextDouble() + 0.1;
                this.qS[s][i] = 1.;
                qSum += this.qS[s][i];
            }
            for (int s = 0; s < this.numMixtures; s++)
                this.qS[s][i] /= qSum;
        }
        for (int d = 0; d < this.numMixtures; d++)
            for (int i = 0; i < this.sampleSize; i++) {
                for (int x = 0; x < this.numLatents; x++) {
                    this.qXMean[d][i][x] = r.nextGaussian();
                    //this.qXMean[d][i][x] = 1.;
                    this.qEX[d][i][x] = this.qXMean[d][i][x];
                    for (int x2 = 0; x2 < this.numLatents; x2++) {
                        if (x == x2)
                            this.qXCov[d][i][x][x2] = 1.;
                            //this.qXCov[d][i][x][x2] = r.nextDouble() + 0.2;
                        else
                            this.qXCov[d][i][x][x2] = 0.;
                    }
                }
                for (int x = 0; x < this.numMainLatents; x++) {
                    for (int x2 = 0; x2 < this.numMainLatents; x2++) {
                        if (x == x2)
                            this.qMainXCov[d][i][x][x2] = this.qXCov[d][i][x][x2];
                        else
                            this.qMainXCov[d][i][x][x2] = 0.;
                    }
                }
                for (int x = 0; x < this.numLatents; x++)
                    for (int x2 = 0; x2 < this.numLatents; x2++)
                        this.qEXX[d][i][x][x2] = this.qXCov[d][i][x][x2] + this.qEX[d][i][x] * this.qEX[d][i][x2];
            }
        for (int d = 0; d < this.numMixtures; d++)
            for (int x = 0; x < this.numLatents; x++) {
                for (int p = 0; p < this.xxParents[x].length + this.xyParents[x].length; p++) {
                    this.qBetaMuMean[d][x][p] = r.nextGaussian();
                    /*this.qBetaMuMean[d][x][p] = r.nextGaussian();
                    if (this.qBetaMuMean[d][x][p] <= 0. && this.qBetaMuMean[d][x][p] > -0.5)
                        this.qBetaMuMean[d][x][p] = -0.5;
                    else if (this.qBetaMuMean[d][x][p] > 0. && this.qBetaMuMean[d][x][p] < 0.5)
                        this.qBetaMuMean[d][x][p] = 0.5;*/
                    this.qEBM[d][x][p] = this.qBetaMuMean[d][x][p];
                    for (int p2 = 0; p2 < this.xxParents[x].length + this.xyParents[x].length; p2++) {
                        if (p == p2)
                           //this.qBetaMuCov[d][x][p][p2] = r.nextDouble() + 0.2;
                           this.qBetaMuCov[d][x][p][p2] = 1.;
                        else
                           this.qBetaMuCov[d][x][p][p2] = 0.;
                    }
                }
                for (int p = 0; p < this.xxParents[x].length + this.xyParents[x].length; p++)
                    for (int p2 = 0; p2 < this.xxParents[x].length + this.xyParents[x].length; p2++)
                        this.qEBMBM[d][x][p][p2] = this.qBetaMuCov[d][x][p][p2] +
                                                   this.qEBM[d][x][p] * this.qEBM[d][x][p2];
                this.upsilon[d][x] = 1.;
                //this.upsilon[d][x] = r.nextDouble();
            }
        for (int d = 0; d < this.numMixtures; d++)
            for (int y = 0; y < this.numVars; y++) {
                for (int p = 0; p < this.qLambdaMean[d][y].length; p++) {
                    this.qLambdaMean[d][y][p] = r.nextGaussian();;
                    /*this.qLambdaMean[d][y][p] = r.nextGaussian();
                    if (this.qLambdaMean[d][y][p] <= 0.  && this.qLambdaMean[d][y][p] > -0.5)
                        this.qLambdaMean[d][y][p] = -0.5;
                    else if (this.qLambdaMean[d][y][p] > 0. && this.qLambdaMean[d][y][p] < 0.5)
                        this.qLambdaMean[d][y][p] = 0.5;*/
                    this.qELambda[d][y][p] = this.qLambdaMean[d][y][p];
                    for (int p2 = 0; p2 < this.qLambdaMean[d][y].length; p2++) {
                        if (p == p2)
                            //this.qLambdaCov[d][y][p][p2] = r.nextDouble() + 0.2;
                            this.qLambdaCov[d][y][p][p2] = 1.;
                        else
                            this.qLambdaCov[d][y][p][p2] = 0.;
                    }
                }
                for (int p = 0; p < this.qLambdaMean[d][y].length; p++)
                    for (int p2 = 0; p2 < this.qLambdaMean[d][y].length; p2++)
                        this.qELambdaLambda[d][y][p][p2] = this.qLambdaCov[d][y][p][p2] +
                                                           this.qELambda[d][y][p] * this.qELambda[d][y][p2];
            }
        //this.invPsiA = r.nextDouble(); this.invPsiB = r.nextDouble();
        //this.invPsiA = 0.; this.invPsiB = 0.;
        for (int y = 0; y < this.numVars; y++) {
            //this.qInvPsiA[y] = r.nextDouble(); this.qInvPsiB[y] = r.nextDouble();
            //this.qEInvPsi[y] = this.qInvPsiA[y] / this.qInvPsiB[y];
            this.qEInvPsi[y] = r.nextDouble() + 0.2;
            //this.qEInvPsi[y] = 10.;
        }
        for (int s = 0; s < this.numMixtures; s++) {
            this.sumS[s] = 0.;
            for (int i = 0; i < this.sampleSize; i++)
                    this.sumS[s] += this.qS[s][i];
        }
        computeLikelihoodKernels();
    }

    //Update the variational distribution q(Pi).
    protected void updateQPi() {
        for (int s = 0; s < this.numMixtures; s++) {
            this.qAlphaM[s] = this.alphaMStar[s];
            for (int i = 0; i < this.sampleSize; i++)
                this.qAlphaM[s] += this.qS[s][i];
        }
    }

    //Update the variational parameters upsilon
    protected void updateQUpsilon() {
        for (int d = 0; d < this.numMixtures; d++) {
            double sumQS = 0.;
            for (int i = 0; i < this.sampleSize; i++)
                sumQS += this.qS[d][i];
            for (int l = 0; l < this.numMainLatents; l++) {
                double dataContrib = 0;
                for (int i = 0; i < this.sampleSize; i++) {
                    double contr = this.qEXX[d][i][l][l];
                    for (int p1 = 0; p1 < this.xxParents[l].length; p1++) {
                        contr -= 2. * this.qEBM[d][l][p1] *
                                      this.qEXX[d][i][l][this.xxParents[l][p1]];
                        for (int p2 = 0; p2 < this.xxParents[l].length; p2++)
                            contr += this.qEBMBM[d][l][p1][p2] *
                                     this.qEXX[d][i][this.xxParents[l][p1]][this.xxParents[l][p2]];
                        for (int p2 = 0; p2 < this.xyParents[l].length; p2++)
                            contr += 2 * this.qEBMBM[d][l][p1][this.xxParents[l].length + p2] *
                                         this.qEX[d][i][this.xxParents[l][p1]] * this.rawData[i][this.xyParents[l][p2]];
                    }
                    for (int p1 = 0; p1 < this.xyParents[l].length; p1++) {
                        contr -= 2. * this.qEBM[d][l][this.xxParents[l].length + p1] *
                                            this.qEX[d][i][l] * this.rawData[i][this.xyParents[l][p1]];
                        for (int p2 = 0; p2 < this.xyParents[l].length; p2++)
                            contr += this.qEBMBM[d][l][this.xxParents[l].length + p1][this.xxParents[l].length + p2] *
                                           this.rawData[i][this.xyParents[l][p1]] * this.rawData[i][this.xyParents[l][p2]];
                    }
                    dataContrib += this.qS[d][i] * contr;
                }
                if (dataContrib > 0.)
                    this.upsilon[d][l] = sumQS / dataContrib;
                else
                    this.upsilon[d][l] = Double.MAX_VALUE;
            }
        }
    }

    //Update the variational distribution q(Beta, Mu)
    protected void updateQBetaMu() {
        for (int d = 0; d < this.numMixtures; d++)
            for (int l = 0; l < this.numLatents; l++) {
                if (this.xxParents[l].length + this.xyParents[l].length == 0)
                    continue;
                int numParents = this.xxParents[l].length + this.xyParents[l].length;
                double auxS[][] = new double[numParents][numParents];
                double auxMean[] = new double[numParents];
                for (int i = 0; i < auxS.length; i++) {
                    auxMean[i] = 0.;
                    for (int j = 0; j < auxS.length; j++)
                        auxS[i][j] = 0.;
                }
                for (int i = 0; i < this.sampleSize; i++) {
                    double constant = this.qS[d][i] * this.upsilon[d][l];
                    for (int p1 = 0; p1 < this.xxParents[l].length; p1++) {
                        for (int p2 = p1; p2 < this.xxParents[l].length; p2++)
                            auxS[p1][p2] += constant *
                                            this.qEXX[d][i][this.xxParents[l][p1]][this.xxParents[l][p2]];
                        for (int p2 = 0; p2 < this.xyParents[l].length; p2++)
                            auxS[p1][this.xxParents[l].length + p2] += constant *
                                       this.qEX[d][i][this.xxParents[l][p1]] * this.rawData[i][this.xyParents[l][p2]];
                        auxMean[p1] += constant * this.qEXX[d][i][l][this.xxParents[l][p1]];
                    }
                    for (int p1 = 0; p1 < this.xyParents[l].length; p1++) {
                        for (int p2 = p1; p2 < this.xyParents[l].length; p2++)
                            auxS[this.xxParents[l].length + p1][this.xxParents[l].length + p2] += constant *
                                        this.rawData[i][this.xyParents[l][p1]] * this.rawData[i][this.xyParents[l][p2]];
                        auxMean[this.xxParents[l].length + p1] += constant * this.qEX[d][i][l]*
                                                                  this.rawData[i][this.xyParents[l][p1]];
                    }
                }
                if (this.useBetaPriors)
                    for (int p = 0; p < numParents; p++)
                        auxS[p][p] += this.betaMuInvCovPrior[d][l][p][p];
                for (int p1 = 0; p1 < numParents; p1++)
                    for (int p2 = p1 + 1; p2 < numParents; p2++)
                        auxS[p2][p1] = auxS[p1][p2];
                this.qBetaMuCov[d][l] = MatrixUtils.inverse(auxS);
                this.qBetaMuMean[d][l] = MatrixUtils.product(this.qBetaMuCov[d][l], auxMean);
                this.qEBM[d][l] = this.qBetaMuMean[d][l];
                for (int v1 = 0; v1 < numParents; v1++)
                   for (int v2 = 0; v2 < numParents; v2++)
                       this.qEBMBM[d][l][v1][v2] = this.qBetaMuCov[d][l][v1][v2] + this.qEBM[d][l][v1] * this.qEBM[d][l][v2];
            }
    }

    public void setLambdaPriors(List prior) {
        this.givenPriors = prior;
        //this.updatePriors = false;
    }

    /**
     * Update the variational distribution of the edges (and intercept) into an observed variable y
     */

    protected void updateQLambda() {
        for (int d = 0; d < this.numMixtures; d++)
            for (int v = 0; v < this.numVars; v++) {
                int totalParents = this.yyParents[v].length + this.yxParents[v].length;
                double auxS[][] = new double[totalParents][totalParents];
                double auxMean[] = new double[totalParents];
                for (int i = 0; i < auxS.length; i++) {
                    auxMean[i] = 0.;
                    for (int j = 0; j < auxS.length; j++)
                        auxS[i][j] = 0.;
                }
                for (int i = 0; i < this.sampleSize; i++) {
                    double constant = this.qS[d][i] * this.qEInvPsi[v];
                    for (int p1 = 0; p1 < this.yyParents[v].length; p1++) {
                        for (int p2 = p1; p2 < this.yyParents[v].length; p2++)
                            auxS[p1][p2] += constant *
                                            this.rawData[i][this.yyParents[v][p1]] * this.rawData[i][this.yyParents[v][p2]];
                        for (int p2 = 0; p2 < this.yxParents[v].length; p2++)
                            auxS[p1][this.yyParents[v].length + p2] += constant *
                                       this.rawData[i][this.yyParents[v][p1]] * this.qEX[d][i][this.yxParents[v][p2]];
                        auxMean[p1] += constant * this.rawData[i][v] * this.rawData[i][this.yyParents[v][p1]];
                    }
                    for (int p1 = 0; p1 < this.yxParents[v].length; p1++) {
                        for (int p2 = p1; p2 < this.yxParents[v].length; p2++)
                            auxS[this.yyParents[v].length + p1][this.yyParents[v].length + p2] += constant *
                                    this.qEXX[d][i][this.yxParents[v][p1]][this.yxParents[v][p2]];
                        auxMean[this.yyParents[v].length + p1] +=
                                constant * this.rawData[i][v] * this.qEX[d][i][this.yxParents[v][p1]];
                    }
                }
                for (int t = 0; t < auxS.length; t++)
                     auxS[t][t] += this.lambdaInvCovPrior[d][v][t][t];
                for (int p1 = 0; p1 < totalParents; p1++)
                    for (int p2 = p1 + 1; p2 < totalParents; p2++)
                        auxS[p2][p1] = auxS[p1][p2];
                this.qLambdaCov[d][v] = MatrixUtils.inverse(auxS);
                this.qLambdaMean[d][v] = MatrixUtils.product(this.qLambdaCov[d][v], auxMean);
                this.qELambda[d][v] = this.qLambdaMean[d][v];
                for (int v1 = 0; v1 < totalParents; v1++)
                    for (int v2 = 0; v2 < totalParents; v2++)
                        this.qELambdaLambda[d][v][v1][v2] = this.qLambdaCov[d][v][v1][v2] +
                                                            this.qLambdaMean[d][v][v1] * this.qLambdaMean[d][v][v2];
            }
        computeLikelihoodKernels();
    }

    protected void updateQX() {
        if (this.numLatents == 0)
            return;
        double auxS[][] = new double[this.numLatents][this.numLatents];
        double auxM[] = new double[this.numLatents];
        for (int d = 0; d < this.numMixtures; d++) {
            for (int i = 0; i < this.sampleSize; i++) {
                computePX(i, d);
                double priorIntercept[] = this.xMeanPrior[d];
                for (int x = 0; x < this.numLatents; x++)
                    auxM[x] = 0.;
                for (int v1 = 0; v1 < this.numLatents; v1++)
                    for (int v2 = 0; v2 < this.numLatents; v2++)
                        auxS[v1][v2] = 0.;
                for (int y = 0; y < this.numVars; y++) {
                    for (int c1 = 0; c1 < this.yxParents[y].length; c1++) {
                        auxM[this.yxParents[y][c1]] += this.qEInvPsi[y] * this.rawData[i][y] *
                                                       this.qELambda[d][y][this.yyParents[y].length + c1];
                        for (int c2 = 0; c2 < this.yyParents[y].length; c2++)
                            auxM[this.yxParents[y][c1]] -= this.qEInvPsi[y] *
                                    this.rawData[i][this.yyParents[y][c2]] *
                                    this.qELambdaLambda[d][y][c2][this.yyParents[y].length + c1];
                        auxS[this.yxParents[y][c1]][this.yxParents[y][c1]] += this.qEInvPsi[y] *
                                this.qELambdaLambda[d][y][this.yyParents[y].length + c1][this.yyParents[y].length + c1];
                        for (int c2 = c1 + 1; c2 < this.yxParents[y].length; c2++) {
                            double aux = this.qEInvPsi[y] *
                                         this.qELambdaLambda[d][y]
                                                            [this.yyParents[y].length + c1][this.yyParents[y].length + c2];
                            if (this.yxParents[y][c1] < this.yxParents[y][c2])
                                auxS[this.yxParents[y][c1]][this.yxParents[y][c2]] += aux;
                            else
                                auxS[this.yxParents[y][c2]][this.yxParents[y][c1]] += aux;
                        }
                    }
                }
                for (int v1 = 0; v1 < this.numLatents; v1++) {
                    auxM[v1] += priorIntercept[v1];
                    for (int v2 = v1; v2 < this.numLatents; v2++)
                         auxS[v1][v2] += this.xInvCovPrior[d][v1][v2];
                }
                for (int v1 = 0; v1 < this.numLatents - 1; v1++)
                     for (int v2 = v1 + 1; v2 < this.numLatents; v2++)
                         auxS[v2][v1] = auxS[v1][v2];
                double auxSInv[][] = MatrixUtils.inverse(auxS);
                this.qXMean[d][i] = MatrixUtils.product(auxSInv, auxM);
                this.qEX[d][i] = this.qXMean[d][i];
                this.qXCov[d][i] = auxSInv;
                for (int x1 = 0; x1 < this.numLatents; x1++)
                    for (int x2 = x1; x2 < this.numLatents; x2++)
                        this.qEXX[d][i][x1][x2] = this.qEXX[d][i][x2][x1] = this.qXCov[d][i][x1][x2] +
                                                                            this.qEX[d][i][x1] * this.qEX[d][i][x2];
                for (int x1 = 0; x1 < this.numMainLatents; x1++)
                    for (int x2 = 0; x2 < this.numMainLatents; x2++)
                        this.qMainXCov[d][i][x1][x2] = this.qXCov[d][i][x1][x2];
            }
        }
        computeLikelihoodKernels();
    }

    protected void computePX(int i, int d) {
        double m[] = new double[this.numLatents];
        double S[][] = new double[this.numLatents][this.numLatents];
        for (int x = 0; x < this.numLatents; x++) {
            m[x] = 0.;
            for (int y = 0; y < this.xyParents[x].length; y++)
                m[x] += this.upsilon[d][x] * this.qEBM[d][x][this.xxParents[x].length + y] *
                                             this.rawData[i][this.xyParents[x][y]];
        }
        for (int v1 = 0; v1 < this.numLatents; v1++)
            for (int v2 = 0; v2 < this.numLatents; v2++)
                S[v1][v2] = 0.;
        for (int x = 0; x < this.numLatents; x++) {
            S[x][x] += this.upsilon[d][x];
            for (int c = 0; c < this.xxParents[x].length; c++) {
                for (int y = 0; y < this.xyParents[x].length; y++)
                    m[this.xxParents[x][c]] -= this.upsilon[d][x] *
                                               this.qEBMBM[d][x][c][this.xxParents[x].length + y] *
                                               this.rawData[i][this.xyParents[x][y]];
                S[this.xxParents[x][c]][this.xxParents[x][c]] +=
                        this.upsilon[d][x] * this.qEBMBM[d][x][c][c];
                double aux = this.upsilon[d][x] * this.qEBM[d][x][c];
                if (x < this.xxParents[x][c])
                    S[x][this.xxParents[x][c]] -= aux;
                else
                    S[this.xxParents[x][c]][x] -= aux;
            }
            for (int c1 = 0; c1 < this.xxParents[x].length - 1; c1++)
                 for (int c2 = c1 + 1; c2 < this.xxParents[x].length; c2++) {
                     double aux = this.upsilon[d][x] * this.qEBMBM[d][x][c1][c2];
                     if (this.xxParents[x][c1] < this.xxParents[x][c2])
                         S[this.xxParents[x][c1]][this.xxParents[x][c2]] += aux;
                     else
                         S[this.xxParents[x][c2]][this.xxParents[x][c1]] += aux;
                 }
        }
        for (int v1 = 0; v1 < this.numLatents - 1; v1++)
             for (int v2 = v1 + 1; v2 < this.numLatents; v2++)
                 S[v2][v1] = S[v1][v2];
        this.xInvCovPrior[d] = S;
        //this.xMeanPrior[d] = MatrixUtils.product(MatrixUtils.inverseGj(this.xInvCovPrior[d], this.numLatents), m);
        this.xMeanPrior[d] = m;
    }

    protected void updateQS() {
        if (this.numMixtures == 1)
            return;
        double startS[] = new double[this.numMixtures];
        double sumAlpha = 0.;
        for (int s = 0; s < this.numMixtures; s++) {
            sumAlpha += this.qAlphaM[s];
            this.sumS[s] = 0.;
        }
        if (this.usePriors)
           for (int s = 0; s < this.numMixtures; s++)
               startS[s] = RBASProbUtils.digamma(this.qAlphaM[s]) - RBASProbUtils.digamma(sumAlpha);
        else
           for (int s = 0; s < this.numMixtures; s++)
               startS[s] = 0.;
        for (int i = 0; i < this.sampleSize; i++) {
            double sumQs = 0., maxQ = -Double.MAX_VALUE;
            for (int s = 0; s < this.numMixtures; s++) {
                this.qS[s][i] = startS[s];
                if (this.numLatents > 0)
                   this.qS[s][i] += expectedLnPX(i, s) + 0.5 * Math.log(MatrixUtils.determinant(this.qXCov[s][i]));
                for (int y = 0; y < this.numVars; y++) {
                    this.qS[s][i] -= 0.5 * this.qEInvPsi[y] * this.likelihoodKernels[s][i][y];
                    //this.qS[s][i] += 0.5 * (RBASProbUtils.digamma(this.qInvPsiA[y]) - Math.log(this.qInvPsiB[y]));
                    this.qS[s][i] += 0.5 * Math.log(this.qEInvPsi[y]);
                }
                if (this.qS[s][i] > maxQ)
                    maxQ = this.qS[s][i];
            }
            for (int s = 0; s < this.numMixtures; s++) {
                this.qS[s][i] = Math.exp(this.qS[s][i] - maxQ);
                sumQs += this.qS[s][i];
            }
            for (int s = 0; s < this.numMixtures; s++) {
                this.qS[s][i] /= sumQs;
                this.sumS[s] += this.qS[s][i];
            }
        }

    }

    protected void computeLikelihoodKernels() {
        mainLoop:
        for (int d = 0; d < this.numMixtures; d++)
            for (int i = 0; i < this.sampleSize; i++)
                for (int y = 0; y < this.numVars; y++) {

                    double varYOthers = 0.;
                    for (int y2 = 0; y2 < this.yyParents[y].length; y2++)
                        varYOthers += this.qELambda[d][y][y2] * this.rawData[i][y] * this.rawData[i][this.yyParents[y][y2]];
                    for (int x = 0; x < this.yxParents[y].length; x++)
                        varYOthers += this.qELambda[d][y][this.yyParents[y].length + x] *
                                      this.rawData[i][y] * this.qEX[d][i][this.yxParents[y][x]];
                    this.likelihoodKernels[d][i][y] = this.rawData[i][y] * this.rawData[i][y]  - 2. * varYOthers;

                    double varYpOthers = 0.;
                    for (int y2 = 0; y2 < this.yyParents[y].length; y2++)
                        for (int y3 = 0; y3 < this.yyParents[y].length; y3++)
                             varYpOthers += this.qELambdaLambda[d][y][y2][y3] *
                                            this.rawData[i][this.yyParents[y][y2]] *
                                            this.rawData[i][this.yyParents[y][y3]];
                    for (int y2 = 0; y2 < this.yyParents[y].length; y2++)
                        for (int x = 0; x < this.yxParents[y].length; x++)
                             varYpOthers += 2 * this.qELambdaLambda[d][y][y2][this.yyParents[y].length + x] *
                                            this.rawData[i][this.yyParents[y][y2]] *
                                            this.qEX[d][i][this.yxParents[y][x]];
                    this.likelihoodKernels[d][i][y] += varYpOthers;

                    double varXpOthers = 0.;
                    for (int x = 0; x < this.yxParents[y].length; x++) {
                        for (int x2 = 0; x2 < this.yxParents[y].length; x2++)
                             varXpOthers += this.qELambdaLambda[d][y][this.yyParents[y].length + x][this.yyParents[y].length + x2] *
                                            this.qEXX[d][i][this.yxParents[y][x]][this.yxParents[y][x2]];
                    }
                    this.likelihoodKernels[d][i][y] += varXpOthers;

                    if (this.likelihoodKernels[d][i][y] < 0.) {
                        this.converging = false;
                        System.out.println("No convergence " + this.likelihoodKernels[d][i][y]);
                        break mainLoop;
                    }
                }
    }

    protected void updateQPsi() {
        for (int y = 0; y < this.numVars; y++) {
            double dataContrib = 0.;
            for (int d = 0; d < this.numMixtures; d++)
                for (int i = 0; i < this.sampleSize; i++)
                    dataContrib += this.qS[d][i] * this.likelihoodKernels[d][i][y];
            if (dataContrib < 0.) {
                this.converging = false;
                return;
            }
            //this.qInvPsiA[y] = this.invPsiA + 0.5 * this.sampleSize;
            //this.qInvPsiB[y] = this.invPsiB + 0.5 * dataContrib;
            //this.qEInvPsi[y] = this.qInvPsiA[y] / this.qInvPsiB[y];
            this.qEInvPsi[y] = this.sampleSize / dataContrib;
        }
    }

    /******************************************************************************************************************
     * Score function evaluation
     */

    protected double fEval() {
        double f = 0.;
        //Penalty term: KL-divergence between priors and approximate posteriors
        if (this.usePriors) {
            f = -RBASProbUtils.klDivergenceDirichlet(this.qAlphaM, this.alphaMStar);
            for (int d = 0; d < this.numMixtures; d++) {
                for (int l = 0; l < this.numLatents; l++) {
                    if (this.xxParents[l].length + this.xyParents[l].length == 0 || !this.useBetaPriors)
                        continue;
                    f -= RBASProbUtils.klDivergenceNormal(this.qBetaMuMean[d][l], this.qBetaMuCov[d][l],
                                                      this.betaMuMeanPrior[d][l], this.betaMuCovPrior[d][l]);
                }
                for (int y = 0; y < this.numVars; y++)
                    f -= RBASProbUtils.klDivergenceNormal(this.qLambdaMean[d][y], this.qLambdaCov[d][y],
                                                      this.lambdaMeanPrior[d][y], this.lambdaCovPrior[d][y]);
            }
            //for (int y = 0; y < this.numVars; y++)
            //    f -= RBASProbUtils.klDivergenceGamma(this.qInvPsiA[y], this.qInvPsiB[y], this.invPsiA, this.invPsiB);
        } else
            f -= 0.5 * Math.log(this.sampleSize) * this.numParameters;
        //Fitness term: likelihood averaged over approximate posteriors
        double alphaMTotal = 0;
        for (int i = 0; i < this.qAlphaM.length; i++)
            alphaMTotal += this.qAlphaM[i];
        for (int i = 0; i < this.sampleSize; i++)
            for (int d = 0; d < this.numMixtures; d++) {
                if (this.qS[d][i] == 0.)
                    continue;
                double partial;
                if (this.usePriors)
                    partial = RBASProbUtils.digamma(this.qAlphaM[d]) - RBASProbUtils.digamma(alphaMTotal)
                                 - Math.log(this.qS[d][i]);
                else
                    partial = 0.;
                if (this.numLatents > 0)
                    partial += expectedLnPX(i, d) + RBASProbUtils.gaussianEntropy(this.qXCov[d][i]);
                for (int y = 0; y < this.numVars; y++)
                    partial -= 0.5 * this.qEInvPsi[y] * this.likelihoodKernels[d][i][y];
                f += this.qS[d][i] * partial;
            }
        double logDetPsi = 0.;
        for (int y = 0; y < this.numVars; y++)
            //logDetPsi += RBASProbUtils.digamma(this.qInvPsiA[y]) - Math.log(this.qInvPsiB[y]);
            logDetPsi += Math.log(this.qEInvPsi[y]);
        f += 0.5 * logDetPsi * this.sampleSize;
        return f;
    }

    public void fEvalDebug() {
        //Penalty term: KL-divergence between priors and approximate posteriors
        System.out.println("NumVars = " + this.numVars);
        double f = 0.;
        double penaltyBeta = 0., penaltyLambda = 0.;
        if (this.usePriors) {
            f = -RBASProbUtils.klDivergenceDirichlet(this.qAlphaM, this.alphaMStar);
            //System.out.println("fEvalDebug.klDirichlet = " + f);
            for (int d = 0; d < this.numMixtures; d++) {
                for (int l = 0; l < this.numLatents; l++) {
                    if (this.xxParents[l].length + this.xyParents[l].length == 0 || !this.useBetaPriors)
                        continue;
                    f -= RBASProbUtils.klDivergenceNormal(this.qBetaMuMean[d][l], this.qBetaMuCov[d][l],
                                                      this.betaMuMeanPrior[d][l], this.betaMuCovPrior[d][l]);
                    penaltyBeta -= RBASProbUtils.klDivergenceNormal(this.qBetaMuMean[d][l], this.qBetaMuCov[d][l],
                                                                this.betaMuMeanPrior[d][l], this.betaMuCovPrior[d][l]);
                }
                for (int y = 0; y < this.numVars; y++) {
                    f -= RBASProbUtils.klDivergenceNormal(this.qLambdaMean[d][y], this.qLambdaCov[d][y],
                                                      this.lambdaMeanPrior[d][y], this.lambdaCovPrior[d][y]);
                    penaltyLambda -= RBASProbUtils.klDivergenceNormal(this.qLambdaMean[d][y], this.qLambdaCov[d][y],
                                                      this.lambdaMeanPrior[d][y], this.lambdaCovPrior[d][y]);
                    System.out.println("..Penalty for variable " + this.indicators[y].toString() + " = " +
                                      (-RBASProbUtils.klDivergenceNormal(this.qLambdaMean[d][y], this.qLambdaCov[d][y],
                                                      this.lambdaMeanPrior[d][y], this.lambdaCovPrior[d][y])));
                    /*for (int i = 0; i < this.lambdaMeanPrior[d][y].length; i++) {
                        System.out.println("   ..Mean penalty: " + this.qLambdaMean[d][y][i] + " " +
                                                          this.lambdaMeanPrior[d][y][i]);
                        for (int j = 0; j < this.lambdaMeanPrior[d][y].length; j++)
                        System.out.println("      ..Covariance penalty: " + this.qLambdaCov[d][y][i][j] + " " +
                                                          this.lambdaCovPrior[d][y][i][j]);
                    }*/
                }
            }
        }
        //for (int y = 0; y < this.numVars; y++)
        //    f -= RBASProbUtils.klDivergenceGamma(this.qInvPsiA[y], this.qInvPsiB[y], this.invPsiA, this.invPsiB);
        //System.out.println("fEvalDebug.otherPenalties = " + f);
        double fPenalty = f;

        //Fitness term: likelihood averaged over approximate posteriors

        double alphaMTotal = 0;
        for (int i = 0; i < this.qAlphaM.length; i++)
            alphaMTotal += this.qAlphaM[i];
        for (int i = 0; i < this.sampleSize; i++)
            for (int d = 0; d < this.numMixtures; d++) {
                if (this.qS[d][i] == 0.)
                    continue;
                double partial;
                if (this.usePriors)
                    partial = RBASProbUtils.digamma(this.qAlphaM[d]) - RBASProbUtils.digamma(alphaMTotal)
                                 - Math.log(this.qS[d][i]);
                else
                    partial = 0.;
                if (this.numLatents > 0)
                    partial += expectedLnPX(i, d) + RBASProbUtils.gaussianEntropy(this.qXCov[d][i]);
                for (int y = 0; y < this.numVars; y++)
                    partial -= 0.5 * this.qEInvPsi[y] * this.likelihoodKernels[d][i][y];
                f += this.qS[d][i] * partial;
                //System.out.println(i + ". " + partial);
                //System.out.println(i + ". " + f);
            }
        //System.out.println("fEvalDebug.withLikelihood = " + f);
        double logDetPsi = 0.;
        for (int y = 0; y < this.numVars; y++)
            //logDetPsi += RBASProbUtils.digamma(this.qInvPsiA[y]) - Math.log(this.qInvPsiB[y]);
            logDetPsi += Math.log(this.qEInvPsi[y]);
        f += 0.5 * logDetPsi * this.sampleSize;
        //System.out.println("fEvalDebug.final = " + f);
        System.out.println("fEvalDebug.penalty = " + fPenalty);
        System.out.println("fEvalDebug.penaltyBeta = " + penaltyBeta);
        System.out.println("fEvalDebug.penaltyLambda = " + penaltyLambda);
        System.out.println("fEvalDebug.likelihoodOnly = " + (f - fPenalty));
    }

    protected double expectedLnPX(int i, int d) {
        double result = -0.5 * (this.numLatents) * Math.log(2. * Math.PI);
        for (int x = 0; x < this.numLatents; x++) {
            result += 0.5 * Math.log(this.upsilon[d][x]) - 0.5 * this.upsilon[d][x] * this.qEXX[d][i][x][x];
            for (int p1 = 0; p1 < this.xxParents[x].length; p1++) {
                result += this.upsilon[d][x] * this.qEBM[d][x][p1] * this.qEXX[d][i][x][this.xxParents[x][p1]];
                for (int p2 = 0; p2 < this.xxParents[x].length; p2++)
                    result -= 0.5 * this.upsilon[d][x] * this.qEBMBM[d][x][p1][p2] *
                              this.qEXX[d][i][this.xxParents[x][p1]][this.xxParents[x][p2]];
                for (int p2 = 0; p2 < this.xyParents[x].length; p2++)
                    result -= this.upsilon[d][x] * this.qEBMBM[d][x][p1][this.xxParents[x].length + p2] *
                              this.qEX[d][i][this.xxParents[x][p1]] * this.rawData[i][this.xyParents[x][p2]];
            }
            for (int p1 = 0; p1 < this.xyParents[x].length; p1++) {
                result += this.upsilon[d][x] * this.qEBM[d][x][this.xxParents[x].length + p1] *
                          this.qEX[d][i][x] * this.rawData[i][this.xyParents[x][p1]];
                for (int p2 = 0; p2 < this.xyParents[x].length; p2++)
                    result -= 0.5 * this.upsilon[d][x] *
                              this.qEBMBM[d][x][this.xxParents[x].length + p1][this.xxParents[x].length + p2] *
                              this.rawData[i][this.xyParents[x][p1]] * this.rawData[i][this.xyParents[x][p2]];
            }
        }
        return result;
    }

    public double getScore() {
        return this.score;
    }

    /******************************************************************************************************************
     * Hyper-parameter optimization
     */

    /*protected void updatePsiHyperParameters() {
        if (!this.usePriors)
            return;
        double offset1 = 0.;
        for (int y = 0; y < this.numVars; y++)
            offset1 += RBASProbUtils.digamma(this.qInvPsiA[y]) - Math.log(this.qInvPsiB[y]);
        offset1 /= this.numVars;
        double offset2 = 0.;
        for (int y = 0; y < this.numVars; y++)
            offset2 += this.qEInvPsi[y];
        offset2 /= this.numVars;
        double errorA, errorB;
        do {
            double oldA = this.invPsiA, oldB = this.invPsiB;
            this.invPsiA = RBASProbUtils.inverseDigamma(Math.log(this.invPsiB) + offset1);
            this.invPsiB = this.invPsiA / offset2;
            errorA = Math.abs(oldA - this.invPsiA); errorB = Math.abs(oldB - this.invPsiB);
        } while (errorA > 1.e-1 || errorB > 1.e-1);
    }*/

    protected void updateLambdaHyperParameters() {
        if (!this.usePriors || !this.updatePriors)
            return;
        if (this.singleInvCovParam) {
            int numYParams = 0;
            double sumEE = 0.;
            for (int d = 0; d < this.numMixtures; d++)
                for (int y = 0; y < this.numVars; y++) {
                    numYParams += this.yyParents[y].length + this.yxParents[y].length;
                    for (int p = 0; p < this.qELambdaLambda[d][y].length; p++)
                        sumEE += this.qELambdaLambda[d][y][p][p];
                }
            for (int y = 0; y < this.numVars; y++)
                this.lambdaInvCovPriorParam[y] = numYParams / sumEE;
        }
        else {
            for (int y = 0; y < this.numVars; y++) {
                int numYParams = 0;
                double sumEE = 0.;
                for (int d = 0; d < this.numMixtures; d++) {
                    numYParams += this.yyParents[y].length + this.yxParents[y].length;
                    for (int p = 0; p < this.qELambdaLambda[d][y].length; p++)
                        sumEE += this.qELambdaLambda[d][y][p][p];
                }
                this.lambdaInvCovPriorParam[y] = numYParams / sumEE;
            }
        }
        for (int d = 0; d < this.numMixtures; d++)
            for (int y = 0; y < this.numVars; y++) {
                int numParents = this.yyParents[y].length + this.yxParents[y].length;
                for (int v = 0; v < numParents; v++) {
                    this.lambdaInvCovPrior[d][y][v][v] = this.lambdaInvCovPriorParam[y];
                    this.lambdaCovPrior[d][y][v][v] = 1. / this.lambdaInvCovPriorParam[y];
                }
            }
   }

    protected void updateBetaMuHyperParameters() {
        if (!this.usePriors || !this.useBetaPriors)
            return;
        int numXParams = 0;
        double sumEE = 0.;
        for (int d = 0; d < this.numMixtures; d++)
            for (int x = 0; x < this.numLatents; x++) {
                numXParams += this.xxParents[x].length + this.xyParents[x].length;
                for (int p = 0; p < this.qEBMBM[d][x].length; p++)
                    sumEE += this.qEBMBM[d][x][p][p];
            }
        this.betaMuInvCovPriorParam = numXParams / sumEE;
        for (int d = 0; d < this.numMixtures; d++)
            for (int l = 0; l < this.numLatents; l++) {
                for (int v = 0; v < this.xxParents[l].length + this.xyParents[l].length; v++) {
                    this.betaMuInvCovPrior[d][l][v][v] = this.betaMuInvCovPriorParam;
                    this.betaMuCovPrior[d][l][v][v] = 1. / this.betaMuInvCovPriorParam;
                }
            }
   }

    /****************************************************************************************************************
     * STRUCTURAL EM METHODS: GENERAL SEMS
     */

    double oldQLambdaCov[][][], oldQLambdaMean[][], oldQELambda[][], oldQELambdaLambda[][][];
    double oldQInvPsiA, oldQInvPsiB, oldQEInvPsi;
    double oldLambdaMeanPrior[][], oldLambdaInvCovPrior[][][], oldLambdaCovPrior[][][], oldLambdaInvCovPriorParam;
    double oldLikelihoodKernels[][];
    int    newYYParents[], newYXParents[], oldYYParents[], oldYXParents[];

    double oldBetaMuMeanPrior[][], oldBetaMuCovPrior[][][], oldBetaMuInvCovPrior[][][], oldBetaMuInvCovPriorParam;
    double oldQBetaMuMean[][], oldQBetaMuCov[][][];
    double oldQEBM[][], oldQEBMBM[][][], oldupsilon[];
    int    newXYParents[], newXXParents[], oldXYParents[], oldXXParents[];

    private void createStructuralEmVariables() {
        this.oldQLambdaCov = new double[this.numMixtures][][];
        this.oldQLambdaMean = new double[this.numMixtures][];
        this.oldQELambda = new double[this.numMixtures][];
        this.oldQELambdaLambda = new double[this.numMixtures][][];
        this.oldLambdaMeanPrior = new double[this.numMixtures][];
        this.oldLambdaInvCovPrior = new double[this.numMixtures][][];
        this.oldLambdaCovPrior = new double[this.numMixtures][][];
        this.oldLikelihoodKernels = new double[this.numMixtures][this.sampleSize];
        this.oldBetaMuMeanPrior = new double[this.numMixtures][];
        this.oldBetaMuCovPrior = new double[this.numMixtures][][];
        this.oldBetaMuInvCovPrior = new double[this.numMixtures][][];
        this.oldQBetaMuMean = new double[this.numMixtures][];
        this.oldQBetaMuCov = new double[this.numMixtures][][];
        this.oldQEBM = new double[this.numMixtures][];
        this.oldQEBMBM = new double[this.numMixtures][][];
        this.oldupsilon = new double[this.numMixtures];
    }

    public double fastEdgeAdditionEval(Node node1, Node node2) {
        return fastEdgeAddition(node1, node2, false);
    }

    public double commitFastEdgeAddition(Node node1, Node node2) {
        return fastEdgeAddition(node1, node2, true);
    }

    public double fastEdgeDeletionEval(Node node1, Node node2) {
        return fastEdgeDeletion(node1, node2, false);
    }

    public double commitFastEdgeDeletion(Node node1, Node node2) {
        return fastEdgeDeletion(node1, node2, true);
    }

    public double fastEdgeReversalEval(Node node1, Node node2) {
        double score1 = fastEdgeDeletion(node1, node2, false);
        double score2 = fastEdgeAddition(node2, node1, false);
        return score1 + score2;
    }

    public double commitFastEdgeReversal(Node node1, Node node2) {
        double score1 = fastEdgeDeletion(node1, node2, true);
        double score2 = fastEdgeAddition(node2, node1, true);
        return score1 + score2;
    }

    private double fastEdgeAddition(Node node1, Node node2, boolean commit) {
        if (node2.getNodeType() == NodeType.MEASURED) {
            String variable = node2.toString();
            double currentScore = 0.;
            int y = -1;
            for (int i = 0; i < this.numVars; i++) {
                if (this.indicators[i].toString().equals(variable)) {
                    y = i;
                    break;
                }
            }
            if (y < 0)
                throw new RuntimeException("This variable does not exist: " + variable);
            //Backup parameters
            this.oldYYParents = this.yyParents[y];
            this.oldYXParents = this.yxParents[y];
            //Remove old score contribution
            for (int d = 0; d < this.numMixtures; d++) {
                if (this.usePriors)
                    currentScore += RBASProbUtils.klDivergenceNormal(this.qLambdaMean[d][y], this.qLambdaCov[d][y],
                                            this.lambdaMeanPrior[d][y], this.lambdaCovPrior[d][y]);
                else
                    currentScore += 0.5 * Math.log(this.sampleSize) * this.qLambdaMean[d][y].length;
                for (int i = 0; i < this.sampleSize; i++)
                    currentScore += 0.5 * this.qS[d][i] * this.qEInvPsi[y] * this.likelihoodKernels[d][i][y];
            }
            currentScore -= 0.5 * Math.log(this.qEInvPsi[y]) * this.sampleSize;
            //Estimate its new parameters
            if (node1.getNodeType() == NodeType.MEASURED) {
                String variable2 = node1.toString();
                int y2 = -1;
                for (int i = 0; i < this.numVars; i++) {
                    if (this.indicators[i].toString().equals(variable2)) {
                        y2 = i;
                        break;
                    }
                }
                if (y2 < 0)
                    throw new RuntimeException("This variable does not exist: " + variable);
                this.yyParents[y] = new int[this.oldYYParents.length + 1];
                for (int i = 0; i < this.oldYYParents.length - 1; i++)
                    this.yyParents[y][i] = this.oldYYParents[i];
                this.yyParents[y][this.oldYYParents.length - 1] = y2;
                this.yyParents[y][this.oldYYParents.length] = this.numVars;
            } else {
                String variable2 = node1.toString();
                int y2 = -1;
                for (int i = 0; i < this.numLatents; i++) {
                    if (this.latents[i].toString().equals(variable2)) {
                        y2 = i;
                        break;
                    }
                }
                if (y2 < 0)
                    throw new RuntimeException("This variable does not exist: " + variable);
                this.yxParents[y] = new int[this.oldYXParents.length + 1];
                for (int i = 0; i < this.oldYXParents.length; i++)
                    this.yxParents[y][i] = this.oldYXParents[i];
                this.yxParents[y][this.oldYXParents.length] = y2;
            }
            double newExtraScore = singleOptimize(y, true, commit);
            if (!commit) {
                this.yyParents[y] = this.oldYYParents;
                this.yxParents[y] = this.oldYXParents;
            } else
                this.score += currentScore + newExtraScore;
            return currentScore + newExtraScore;
        } else {
            String variable = node2.toString();
            double currentScore = 0.;
            int x = -1;
            for (int i = 0; i < this.numLatents; i++) {
                if (this.latents[i].toString().equals(variable)) {
                    x = i;
                    break;
                }
            }
            if (x < 0)
                throw new RuntimeException("This variable does not exist: " + variable);
            this.oldXXParents = this.xxParents[x];
            this.oldXYParents = this.xyParents[x];
            //Remove its penalty contribution
            for (int d = 0; d < this.numMixtures; d++) {
                if (this.usePriors) {
                    if (this.xxParents[x].length + this.xyParents[x].length != 0 && this.useBetaPriors)
                        currentScore += RBASProbUtils.klDivergenceNormal(this.qBetaMuMean[d][x], this.qBetaMuCov[d][x],
                                                           this.betaMuMeanPrior[d][x], this.betaMuCovPrior[d][x]);
                }
                else
                    currentScore += 0.5 * Math.log(this.sampleSize) * this.qBetaMuMean[d][x].length;
                for (int i = 0; i < this.sampleSize; i++)
                    currentScore -= this.qS[d][i] * expectedLnPX(i, x, d);
            }
            //Estimate its new parameters
            if (node1.getNodeType() == NodeType.MEASURED) {
                String variable2 = node1.toString();
                int y2 = -1;
                for (int i = 0; i < this.numVars; i++) {
                    if (this.indicators[i].toString().equals(variable2)) {
                        y2 = i;
                        break;
                    }
                }
                if (y2 < 0)
                    throw new RuntimeException("This variable does not exist: " + variable);
                this.xyParents[x] = new int[this.oldXYParents.length + 1];
                for (int i = 0; i < this.oldXYParents.length; i++)
                    this.xyParents[x][i] = this.oldXYParents[i];
                this.xyParents[x][this.oldXYParents.length] = y2;
            } else {
                String variable2 = node1.toString();
                int x2 = -1;
                for (int i = 0; i < this.numLatents; i++) {
                    if (this.latents[i].toString().equals(variable2)) {
                        x2 = i;
                        break;
                    }
                }
                if (x2 < 0)
                    throw new RuntimeException("This variable does not exist: " + variable);
                this.xxParents[x] = new int[this.oldXXParents.length + 1];
                for (int i = 0; i < this.oldXXParents.length; i++)
                    this.xxParents[x][i] = this.oldXXParents[i];
                this.xxParents[x][this.oldXXParents.length] = x2;
            }
            double newExtraScore = singleOptimize(x, false, commit);
            if (!commit) {
                this.xyParents[x] = this.oldXYParents;
                this.xxParents[x] = this.oldXXParents;
            } else
                this.score += currentScore + newExtraScore;
            return currentScore + newExtraScore;
        }
    }

    private double fastEdgeDeletion(Node node1, Node node2, boolean commit) {
        if (node2.getNodeType() == NodeType.MEASURED) {
            String variable = node2.toString();
            double currentScore = 0.;
            int y = -1;
            for (int i = 0; i < this.numVars; i++) {
                if (this.indicators[i].toString().equals(variable)) {
                    y = i;
                    break;
                }
            }
            if (y < 0)
                throw new RuntimeException("This variable does not exist: " + variable);
            //Backup parameters
            this.oldYYParents = this.yyParents[y];
            this.oldYXParents = this.yxParents[y];
            //Remove old score contribution
            for (int d = 0; d < this.numMixtures; d++) {
                if (this.usePriors)
                    currentScore += RBASProbUtils.klDivergenceNormal(this.qLambdaMean[d][y], this.qLambdaCov[d][y],
                                            this.lambdaMeanPrior[d][y], this.lambdaCovPrior[d][y]);
                else
                    currentScore += 0.5 * Math.log(this.sampleSize) * this.qLambdaMean[d][y].length;
                for (int i = 0; i < this.sampleSize; i++)
                    currentScore += 0.5 * this.qS[d][i] * this.qEInvPsi[y] * this.likelihoodKernels[d][i][y];
            }
            currentScore -= 0.5 * Math.log(this.qEInvPsi[y]) * this.sampleSize;

            //Estimate its new parameters
            if (node1.getNodeType() == NodeType.MEASURED) {
                String variable2 = node1.toString();
                int y2 = -1;
                for (int i = 0; i < this.numVars; i++) {
                    if (this.indicators[i].toString().equals(variable2)) {
                        y2 = i;
                        break;
                    }
                }
                if (y2 < 0)
                    throw new RuntimeException("This variable does not exist: " + variable);
                this.yyParents[y] = new int[this.oldYYParents.length - 1];
                int count = 0;
                for (int i = 0; i < this.oldYYParents.length; i++)
                    if (this.oldYYParents[i] != y2)
                        this.yyParents[y][count++] = this.oldYYParents[i];
            } else {
                String variable2 = node1.toString();
                int y2 = -1;
                for (int i = 0; i < this.numLatents; i++) {
                    if (this.latents[i].toString().equals(variable2)) {
                        y2 = i;
                        break;
                    }
                }
                if (y2 < 0)
                    throw new RuntimeException("This variable does not exist: " + variable);
                this.yxParents[y] = new int[this.oldYXParents.length - 1];
                int count = 0;
                for (int i = 0; i < this.oldYXParents.length; i++)
                    if (this.oldYXParents[i] != y2)
                        this.yxParents[y][count++] = this.oldYXParents[i];
            }
            double newExtraScore = singleOptimize(y, true, commit);
            if (!commit) {
                this.yyParents[y] = this.oldYYParents;
                this.yxParents[y] = this.oldYXParents;
            } else
                this.score += currentScore + newExtraScore;
            return currentScore + newExtraScore;
        } else {
            String variable = node2.toString();
            //double currentScore = this.score;
            double currentScore = 0.;
            int x = -1;
            for (int i = 0; i < this.numLatents; i++) {
                if (this.latents[i].toString().equals(variable)) {
                    x = i;
                    break;
                }
            }
            if (x < 0)
                throw new RuntimeException("This variable does not exist: " + variable);
            this.oldXXParents = this.xxParents[x];
            this.oldXYParents = this.xyParents[x];
            //Remove its penalty contribution
            for (int d = 0; d < this.numMixtures; d++) {
                if (this.usePriors) {
                    if (this.xxParents[x].length + this.xyParents[x].length != 0 && this.useBetaPriors)
                        currentScore += RBASProbUtils.klDivergenceNormal(this.qBetaMuMean[d][x], this.qBetaMuCov[d][x],
                                                           this.betaMuMeanPrior[d][x], this.betaMuCovPrior[d][x]);
                }
                else
                    currentScore += 0.5 * Math.log(this.sampleSize) * this.qBetaMuMean[d][x].length;
                for (int i = 0; i < this.sampleSize; i++)
                    currentScore -= this.qS[d][i] * expectedLnPX(i, x, d);
            }
            //Estimate its new parameters
            if (node1.getNodeType() == NodeType.MEASURED) {
                String variable2 = node1.toString();
                int y2 = -1;
                for (int i = 0; i < this.numVars; i++) {
                    if (this.indicators[i].toString().equals(variable2)) {
                        y2 = i;
                        break;
                    }
                }
                if (y2 < 0)
                    throw new RuntimeException("This variable does not exist: " + variable);
                this.xyParents[x] = new int[this.oldXYParents.length - 1];
                int count = 0;
                for (int i = 0; i < this.oldXYParents.length; i++)
                    if (this.oldXYParents[i] != y2)
                        this.xyParents[x][count++] = this.oldXYParents[i];
            } else {
                String variable2 = node1.toString();
                int x2 = -1;
                for (int i = 0; i < this.numLatents; i++) {
                    if (this.latents[i].toString().equals(variable2)) {
                        x2 = i;
                        break;
                    }
                }
                if (x2 < 0)
                    throw new RuntimeException("This variable does not exist: " + variable);
                this.xxParents[x] = new int[this.oldXXParents.length - 1];
                int count = 0;
                for (int i = 0; i < this.oldXXParents.length; i++)
                    if (this.oldXXParents[i] != x2)
                       this.xxParents[x][count++] = this.oldXXParents[i];
            }
            double newExtraScore = singleOptimize(x, false, commit);
            if (!commit) {
                this.xyParents[x] = this.oldXYParents;
                this.xxParents[x] = this.oldXXParents;
            } else
                this.score += currentScore + newExtraScore;
            return currentScore + newExtraScore;
        }
    }
    
    private void initializeNewParams(int v, boolean isObserved) {
        int mult;
        if (this.usePriors)
            mult = 1;
        else
            mult = 0;
        if (isObserved) {
            Random r = new Random();
            this.oldQInvPsiA = this.qInvPsiA[v];
            this.oldQInvPsiB = this.qInvPsiB[v];
            this.oldQEInvPsi = this.qEInvPsi[v];
            this.oldLambdaInvCovPriorParam = this.lambdaInvCovPriorParam[v];
            this.oldBetaMuInvCovPriorParam = this.betaMuInvCovPriorParam;
            for (int d = 0; d < this.numMixtures; d++) {
                this.oldQLambdaCov[d] = this.qLambdaCov[d][v];
                this.oldQELambdaLambda[d] = this.qELambdaLambda[d][v];
                this.oldQLambdaMean[d] = this.qLambdaMean[d][v];
                this.oldQELambda[d] = this.qELambda[d][v];
                this.oldLambdaMeanPrior[d] = this.lambdaMeanPrior[d][v];
                this.oldLambdaCovPrior[d] = this.lambdaCovPrior[d][v];
                this.oldLambdaInvCovPrior[d] = this.lambdaInvCovPrior[d][v];
                for (int i = 0; i < this.sampleSize; i++)
                    this.oldLikelihoodKernels[d][i] = this.likelihoodKernels[d][i][v];

                int numParents = this.yyParents[v].length + this.yxParents[v].length;
                this.qLambdaCov[d][v] = new double[numParents][numParents];
                this.qELambdaLambda[d][v] = new double[numParents][numParents];
                this.qLambdaMean[d][v] = new double[numParents];
                this.qELambda[d][v] = new double[numParents];
                this.lambdaMeanPrior[d][v] = new double[numParents];
                this.lambdaInvCovPrior[d][v] = new double[numParents][numParents];

                for (int v1 = 0; v1 < numParents; v1++) {
                    this.lambdaMeanPrior[d][v][v1] = 0.;
                    for (int v2 = 0; v2 < numParents; v2++) {
                        if (v1 == v2)
                            this.lambdaInvCovPrior[d][v][v1][v2] = mult * this.lambdaInvCovPriorParam[v];
                        else
                            this.lambdaInvCovPrior[d][v][v1][v2] = 0.;
                    }
                }
                if (this.usePriors)
                    this.lambdaCovPrior[d][v] = MatrixUtils.inverse(this.lambdaInvCovPrior[d][v]);
                else
                    this.lambdaCovPrior[d][v] = null;
                for (int p = 0; p < this.qLambdaMean[d][v].length; p++) {
                    this.qLambdaMean[d][v][p] = r.nextGaussian();
                    for (int p2 = 0; p2 < this.qLambdaMean[d][v].length; p2++) {
                        if (p == p2)
                            this.qLambdaCov[d][v][p][p2] = 1.;
                        else
                            this.qLambdaCov[d][v][p][p2] = 0.;
                    }
                }
                for (int p = 0; p < this.qLambdaMean[d][v].length; p++)
                    for (int p2 = 0; p2 < this.qLambdaMean[d][v].length; p2++)
                        this.qELambdaLambda[d][v][p][p2] = this.qLambdaCov[d][v][p][p2] +
                                                           this.qELambda[d][v][p] * this.qELambda[d][v][p2];
            }
            this.qInvPsiA[v] = r.nextDouble(); this.qInvPsiB[v] = r.nextDouble();
            this.qEInvPsi[v] = this.qInvPsiA[v] / this.qInvPsiB[v];
            computeLikelihoodKernels(v);
        } else {
            for (int d = 0; d < this.numMixtures; d++) {
                this.oldBetaMuMeanPrior[d] = this.betaMuMeanPrior[d][v];
                this.oldBetaMuCovPrior[d] = this.betaMuCovPrior[d][v];
                this.oldBetaMuInvCovPrior[d] = this.betaMuInvCovPrior[d][v];
                this.oldLambdaInvCovPriorParam = this.lambdaInvCovPriorParam[v];
                this.oldBetaMuInvCovPriorParam = this.betaMuInvCovPriorParam;
                this.oldQBetaMuMean[d] = this.qBetaMuMean[d][v];
                this.oldQBetaMuCov[d] = this.qBetaMuCov[d][v];
                this.oldQEBM[d] = this.qEBM[d][v];
                this.oldQEBMBM[d] = this.qEBMBM[d][v];
                this.oldupsilon[d] = this.upsilon[d][v];

                int numParents = this.xxParents[v].length + this.xyParents[v].length;
                this.betaMuMeanPrior[d][v] = new double[numParents];
                this.betaMuCovPrior[d][v] = new double[numParents][numParents];
                this.betaMuInvCovPrior[d][v] = new double[numParents][numParents];
                this.qBetaMuMean[d][v] = new double[numParents];
                this.qBetaMuCov[d][v] = new double[numParents][numParents];
                this.qEBM[d][v] = new double[numParents];
                this.qEBMBM[d][v] = new double[numParents][numParents];

                for (int v1 = 0; v1 < numParents; v1++) {
                    this.betaMuMeanPrior[d][v][v1] = 0.; //This *has* to be zero.
                    for (int v2 = 0; v2 < numParents; v2++) {
                        if (v1 == v2)
                            this.betaMuCovPrior[d][v][v1][v2] = mult * this.betaMuInvCovPriorParam;
                        else
                            this.betaMuCovPrior[d][v][v1][v2] = 0.;
                    }
                }
                if (this.usePriors) {
                    if (numParents > 0)
                        this.betaMuInvCovPrior[d][v] = MatrixUtils.inverse(this.betaMuCovPrior[d][v]);
                }
                else {
                    this.betaMuInvCovPrior[d][v] = null;/*new double[this.betaMuCovPrior[d][v].length][this.betaMuCovPrior[d][v].length];
                    for (int p1 = 0; p1 < this.betaMuInvCovPrior[d][v].length; p1++)
                        for (int p2 = 0; p2 < this.betaMuInvCovPrior[d][v].length; p2++)
                            this.betaMuInvCovPrior[d][v][p1][p2] = 0.;*/
                }
                Random r = new Random();
                for (int p = 0; p < this.xxParents[v].length + this.xyParents[v].length; p++)
                    this.qBetaMuMean[d][v][p] = this.qEBM[d][v][p] = r.nextGaussian();
                for (int p = 0; p < this.xxParents[v].length + this.xyParents[v].length; p++)
                    for (int p2 = 0; p2 < this.xxParents[v].length + this.xyParents[v].length; p2++) {
                        if (p == p2)
                           this.qBetaMuCov[d][v][p][p2] = 1.;
                        else
                           this.qBetaMuCov[d][v][p][p2] = 0.;
                        this.qEBMBM[d][v][p][p2] = this.qBetaMuCov[d][v][p][p2] + this.qEBM[d][v][p] * this.qEBM[d][v][p2];
                    }
                this.upsilon[d][v] = 1.;
            }
        }
    }

    private void restoreParameters(int v, boolean isObserved) {
        if (isObserved) {
            this.qInvPsiA[v] = this.oldQInvPsiA;
            this.qInvPsiB[v] = this.oldQInvPsiB;
            this.qEInvPsi[v] = this.oldQEInvPsi;
            for (int d = 0; d < this.numMixtures; d++) {
                this.qLambdaCov[d][v] = this.oldQLambdaCov[d];
                this.qELambdaLambda[d][v] = this.oldQELambdaLambda[d];
                this.qLambdaMean[d][v] = this.oldQLambdaMean[d];
                this.qELambda[d][v] = this.oldQELambda[d];
                this.lambdaMeanPrior[d][v] = this.oldLambdaMeanPrior[d];
                this.lambdaCovPrior[d][v] = this.oldLambdaCovPrior[d];
                this.lambdaInvCovPrior[d][v] = this.oldLambdaInvCovPrior[d];
                this.lambdaInvCovPriorParam[v] = this.oldLambdaInvCovPriorParam;
                for (int i = 0; i < this.sampleSize; i++)
                    this.likelihoodKernels[d][i][v] = this.oldLikelihoodKernels[d][i];
            }
        } else {
            for (int d = 0; d < this.numMixtures; d++) {
                this.betaMuMeanPrior[d][v] = this.oldBetaMuMeanPrior[d];
                this.betaMuCovPrior[d][v] = this.oldBetaMuCovPrior[d];
                this.betaMuInvCovPrior[d][v] = this.oldBetaMuInvCovPrior[d];
                this.betaMuInvCovPriorParam = this.oldLambdaInvCovPriorParam;
                this.qBetaMuMean[d][v] = this.oldQBetaMuMean[d];
                this.qBetaMuCov[d][v] = this.oldQBetaMuCov[d];
                this.qEBM[d][v] = this.oldQEBM[d];
                this.qEBMBM[d][v] = this.oldQEBMBM[d];
                this.upsilon[d][v] = this.oldupsilon[d];
            }
        }
    }

    private double singleOptimize(int v, boolean isObserved, boolean commit) {
        initializeNewParams(v, isObserved);
        double f, newF = newSingleFEval(v, isObserved);
        int iter = 0;
        if (isObserved)
            do {
                f = newF;
                updateQLambda(v); updateLambdaHyperParameters(v);
                updateQPsi(v);
                newF = newSingleFEval(v, isObserved);
                iter++;
            } while (newF - f > MIN_TOLERANCE && iter < 5);
        else {
            do {
                f = newF;
                updateQBetaMu(v); updateBetaMuHyperParameters();
                updateQUpsilon(v);
                newF = newSingleFEval(v, isObserved);
                if (Double.isInfinite(newF)) {
                    System.out.println("!");
                    System.exit(0);
                }
                iter++;
            } while (newF - f > MIN_TOLERANCE && iter < 5);
        }
        if (!commit)
            restoreParameters(v, isObserved);
        return newF;
    }

    private double newSingleFEval(int y, boolean isObserved) {
        double f = 0.;
        if (isObserved) {
            //f -= RBASProbUtils.klDivergenceGamma(this.qInvPsiA[y], this.qInvPsiB[y], this.invPsiA, this.invPsiB);
            for (int d = 0; d < this.numMixtures; d++) {
                if (this.usePriors)
                    f -= RBASProbUtils.klDivergenceNormal(this.qLambdaMean[d][y], this.qLambdaCov[d][y],
                                                      this.lambdaMeanPrior[d][y], this.lambdaCovPrior[d][y]);
                else
                    f -= 0.5 * Math.log(this.sampleSize) * this.qLambdaMean[d][y].length;
                for (int i = 0; i < this.sampleSize; i++)
                    f -= 0.5 * this.qS[d][i] * this.qEInvPsi[y] * this.likelihoodKernels[d][i][y];
            }
            f += 0.5 * Math.log(this.qEInvPsi[y]) * this.sampleSize;
        } else {
            if (this.xxParents[y].length + this.xyParents[y].length > 0 && this.useBetaPriors) {
                if (this.useBetaPriors) {
                    if (this.usePriors)
                        for (int d = 0; d < this.numMixtures; d++)
                            f -= RBASProbUtils.klDivergenceNormal(this.qBetaMuMean[d][y], this.qBetaMuCov[d][y],
                                                              this.betaMuMeanPrior[d][y], this.betaMuCovPrior[d][y]);
                    else
                        for (int d = 0; d < this.numMixtures; d++)
                            f -= 0.5 * Math.log(this.sampleSize) * this.qBetaMuMean[d][y].length;
                }
            }
            for (int i = 0; i < this.sampleSize; i++)
                for (int d = 0; d < this.numMixtures; d++)
                    f += this.qS[d][i] * expectedLnPX(i, y, d);
        }
        return f;
    }

    private double expectedLnPX(int i, int x, int d) {
        double result = -0.5 * Math.log(2. * Math.PI);
        result += 0.5 * Math.log(this.upsilon[d][x]) - 0.5 * this.upsilon[d][x] * this.qEXX[d][i][x][x];
        for (int p1 = 0; p1 < this.xxParents[x].length; p1++) {
            result += this.upsilon[d][x] * this.qEBM[d][x][p1] * this.qEXX[d][i][x][this.xxParents[x][p1]];
            for (int p2 = 0; p2 < this.xxParents[x].length; p2++)
                result -= 0.5 * this.upsilon[d][x] * this.qEBMBM[d][x][p1][p2] *
                          this.qEXX[d][i][this.xxParents[x][p1]][this.xxParents[x][p2]];
            for (int p2 = 0; p2 < this.xyParents[x].length; p2++)
                result -= this.upsilon[d][x] * this.qEBMBM[d][x][p1][this.xxParents[x].length + p2] *
                          this.qEX[d][i][this.xxParents[x][p1]] * this.rawData[i][this.xyParents[x][p2]];
        }
        for (int p1 = 0; p1 < this.xyParents[x].length; p1++) {
            result += this.upsilon[d][x] * this.qEBM[d][x][this.xxParents[x].length + p1] *
                      this.qEX[d][i][x] * this.rawData[i][this.xyParents[x][p1]];
            for (int p2 = 0; p2 < this.xyParents[x].length; p2++)
                result -= 0.5 * this.upsilon[d][x] *
                          this.qEBMBM[d][x][this.xxParents[x].length + p1][this.xxParents[x].length + p2] *
                          this.rawData[i][this.xyParents[x][p1]] * this.rawData[i][this.xyParents[x][p2]];
        }
        return result;
    }

    protected void updateQLambda(int v) {
        for (int d = 0; d < this.numMixtures; d++) {
            int totalParents = this.yyParents[v].length + this.yxParents[v].length;
            double auxS[][] = new double[totalParents][totalParents];
            double auxMean[] = new double[totalParents];
            for (int i = 0; i < auxS.length; i++) {
                auxMean[i] = 0.;
                for (int j = 0; j < auxS.length; j++)
                    auxS[i][j] = 0.;
            }
            for (int i = 0; i < this.sampleSize; i++) {
                double constant = this.qS[d][i] * this.qEInvPsi[v], nextTerm;
                for (int p1 = 0; p1 < this.yyParents[v].length; p1++) {
                    for (int p2 = p1; p2 < this.yyParents[v].length; p2++)
                        auxS[p1][p2] += constant *
                                        this.rawData[i][this.yyParents[v][p1]] * this.rawData[i][this.yyParents[v][p2]];
                    for (int p2 = 0; p2 < this.yxParents[v].length; p2++) {
                        nextTerm = constant *
                                   this.rawData[i][this.yyParents[v][p1]] * this.qEX[d][i][this.yxParents[v][p2]];
                        auxS[p1][this.yyParents[v].length + p2] += nextTerm;
                        //auxS[this.yyParents[v].length + p2][p1] += nextTerm;
                    }
                    auxMean[p1] += constant * this.rawData[i][v] * this.rawData[i][this.yyParents[v][p1]];
                }
                for (int p1 = 0; p1 < this.yxParents[v].length; p1++) {
                    for (int p2 = p1; p2 < this.yxParents[v].length; p2++)
                        auxS[this.yyParents[v].length + p1][this.yyParents[v].length + p2] += constant *
                                this.qEXX[d][i][this.yxParents[v][p1]][this.yxParents[v][p2]];
                    auxMean[this.yyParents[v].length + p1] +=
                            constant * this.rawData[i][v] * this.qEX[d][i][this.yxParents[v][p1]];
                }
                for (int p1 = 0; p1 < totalParents; p1++)
                    for (int p2 = p1 + 1; p2 < totalParents; p2++)
                        auxS[p2][p1] = auxS[p1][p2];
                this.qLambdaCov[d][v] = MatrixUtils.inverse(MatrixUtils.sum(auxS, this.lambdaInvCovPrior[d][v]));
                this.qLambdaMean[d][v] = MatrixUtils.product(this.qLambdaCov[d][v], auxMean);
                this.qELambda[d][v] = this.qLambdaMean[d][v];
                for (int v1 = 0; v1 < totalParents; v1++)
                    for (int v2 = 0; v2 < totalParents; v2++)
                        this.qELambdaLambda[d][v][v1][v2] = this.qLambdaCov[d][v][v1][v2] +
                                                         this.qLambdaMean[d][v][v1] * this.qLambdaMean[d][v][v2];
            }
        }
        computeLikelihoodKernels(v);
    }

    protected void updateLambdaHyperParameters(int y) {
        if (!this.usePriors || !this.updatePriors)
            return;
        if (this.singleInvCovParam) {
            int numYParams = 0;
            double sumEE = 0.;
            for (int d = 0; d < this.numMixtures; d++) {
                    numYParams += this.yyParents[y].length + this.yxParents[y].length;
                    for (int p = 0; p < this.qELambdaLambda[d][y].length; p++)
                        sumEE += this.qELambdaLambda[d][y][p][p];
                }
            this.lambdaInvCovPriorParam[y] = numYParams / sumEE;
            for (int d = 0; d < this.numMixtures; d++) {
                int numParents = this.yyParents[y].length + this.yxParents[y].length;
                for (int v = 0; v < numParents; v++) {
                    this.lambdaInvCovPrior[d][y][v][v] = this.lambdaInvCovPriorParam[y];
                    this.lambdaCovPrior[d][y][v][v] = 1. / this.lambdaInvCovPriorParam[y];
                }
            }
        }
   }

    protected void updateQPsi(int y) {
        double dataContrib = 0.;
        for (int d = 0; d < this.numMixtures; d++)
            for (int i = 0; i < this.sampleSize; i++)
                dataContrib += this.qS[d][i] * this.likelihoodKernels[d][i][y];
        //this.qInvPsiA[y] = this.invPsiA + 0.5 * this.sampleSize;
        //this.qInvPsiB[y] = this.invPsiB + 0.5 * dataContrib;
        //this.qEInvPsi[y] = this.qInvPsiA[y] / this.qInvPsiB[y];
        this.qEInvPsi[y] = this.sampleSize / dataContrib;
    }

    protected void computeLikelihoodKernels(int y) {
        for (int d = 0; d < this.numMixtures; d++)
            for (int i = 0; i < this.sampleSize; i++) {
                double varYOthers = 0.;
                for (int y2 = 0; y2 < this.yyParents[y].length; y2++)
                    varYOthers += this.qELambda[d][y][y2] * this.rawData[i][y] * this.rawData[i][this.yyParents[y][y2]];
                for (int x = 0; x < this.yxParents[y].length; x++)
                    varYOthers += this.qELambda[d][y][this.yyParents[y].length + x] *
                                  this.rawData[i][y] * this.qEX[d][i][this.yxParents[y][x]];
                this.likelihoodKernels[d][i][y] = this.rawData[i][y] * this.rawData[i][y]  - 2. * varYOthers;

                double varYpOthers = 0.;
                for (int y2 = 0; y2 < this.yyParents[y].length; y2++)
                    for (int y3 = 0; y3 < this.yyParents[y].length; y3++)
                         varYpOthers += this.qELambdaLambda[d][y][y2][y3] *
                                        this.rawData[i][this.yyParents[y][y2]] *
                                        this.rawData[i][this.yyParents[y][y3]];
                for (int y2 = 0; y2 < this.yyParents[y].length; y2++)
                    for (int x = 0; x < this.yxParents[y].length; x++)
                         varYpOthers += 2 * this.qELambdaLambda[d][y][y2][this.yyParents[y].length + x] *
                                        this.rawData[i][this.yyParents[y][y2]] *
                                        this.qEX[d][i][this.yxParents[y][x]];
                this.likelihoodKernels[d][i][y] += varYpOthers;

                double varXpOthers = 0.;
                for (int x = 0; x < this.yxParents[y].length; x++) {
                    for (int x2 = 0; x2 < this.yxParents[y].length; x2++)
                         varXpOthers += this.qELambdaLambda[d][y][this.yyParents[y].length + x][this.yyParents[y].length + x2] *
                                        this.qEXX[d][i][this.yxParents[y][x]][this.yxParents[y][x2]];
                }
                this.likelihoodKernels[d][i][y] += varXpOthers;
            }
    }

    private void updateQBetaMu(int l) {
        for (int d = 0; d < this.numMixtures; d++) {
            int numParents = this.xxParents[l].length + this.xyParents[l].length;
            if (numParents == 0)
                return;
            double auxS[][] = new double[numParents][numParents];
            double auxMean[] = new double[numParents];
            for (int i = 0; i < auxS.length; i++) {
                auxMean[i] = 0.;
                for (int j = 0; j < auxS.length; j++)
                    auxS[i][j] = 0.;
            }
            for (int i = 0; i < this.sampleSize; i++) {
                double constant = this.qS[d][i] * this.upsilon[d][l], nextTerm;
                for (int p1 = 0; p1 < this.xxParents[l].length; p1++) {
                    for (int p2 = p1; p2 < this.xxParents[l].length; p2++)
                        auxS[p1][p2] += constant *
                                        this.qEXX[d][i][this.xxParents[l][p1]][this.xxParents[l][p2]];
                    for (int p2 = 0; p2 < this.xyParents[l].length; p2++) {
                        nextTerm = constant *
                                   this.qEX[d][i][this.xxParents[l][p1]] * this.rawData[i][this.xyParents[l][p2]];
                        auxS[p1][this.xxParents[l].length + p2] += nextTerm;
                        auxS[this.xxParents[l].length + p2][p1] += nextTerm;
                    }
                    auxMean[p1] += constant * this.qEXX[d][i][l][this.xxParents[l][p1]];
                }
                for (int p1 = 0; p1 < this.xyParents[l].length; p1++) {
                    for (int p2 = p1; p2 < this.xyParents[l].length; p2++)
                        auxS[this.xxParents[l].length + p1][this.xxParents[l].length + p2] += constant *
                                    this.rawData[i][this.xyParents[l][p1]] * this.rawData[i][this.xyParents[l][p2]];
                    auxMean[this.xxParents[l].length + p1] += constant * this.qEX[d][i][l] *
                                                              this.rawData[i][this.xyParents[l][p1]];
                }
            }
            if (this.useBetaPriors)
                for (int p = 0; p < numParents; p++)
                    auxS[p][p] += this.betaMuInvCovPrior[d][l][p][p];
            for (int p1 = 0; p1 < numParents; p1++)
                for (int p2 = p1 + 1; p2 < numParents; p2++)
                    auxS[p2][p1] = auxS[p1][p2];
            this.qBetaMuCov[d][l] = MatrixUtils.inverse(auxS);
            this.qBetaMuMean[d][l] = MatrixUtils.product(this.qBetaMuCov[d][l], auxMean);
            this.qEBM[d][l] = this.qBetaMuMean[d][l];
            for (int v1 = 0; v1 < numParents; v1++)
               for (int v2 = 0; v2 < numParents; v2++)
                   this.qEBMBM[d][l][v1][v2] = this.qBetaMuCov[d][l][v1][v2] + this.qEBM[d][l][v1] * this.qEBM[d][l][v2];
        }
    }

    private void updateQUpsilon(int l) {
        for (int d = 0; d < this.numMixtures; d++) {
            double sumQS = 0.;
            for (int i = 0; i < this.sampleSize; i++)
                sumQS += this.qS[d][i];
            double dataContrib = 0;
            for (int i = 0; i < this.sampleSize; i++) {
                double contr = this.qEXX[d][i][l][l];
                for (int p1 = 0; p1 < this.xxParents[l].length; p1++) {
                    contr -= 2. * this.qEBM[d][l][p1] *
                                  this.qEXX[d][i][l][this.xxParents[l][p1]];
                    for (int p2 = 0; p2 < this.xxParents[l].length; p2++)
                        contr += this.qEBMBM[d][l][p1][p2] *
                                 this.qEXX[d][i][this.xxParents[l][p1]][this.xxParents[l][p2]];
                    for (int p2 = 0; p2 < this.xyParents[l].length; p2++)
                        contr += 2 * this.qEBMBM[d][l][p1][this.xxParents[l].length + p2] *
                                     this.qEX[d][i][this.xxParents[l][p1]] * this.rawData[i][this.xyParents[l][p2]];
                }
                for (int p1 = 0; p1 < this.xyParents[l].length; p1++) {
                    contr -= 2. * this.qEBM[d][l][this.xxParents[l].length + p1] *
                                        this.qEX[d][i][l] * this.rawData[i][this.xyParents[l][p1]];
                    for (int p2 = 0; p2 < this.xyParents[l].length; p2++)
                        contr += this.qEBMBM[d][l][this.xxParents[l].length + p1][this.xxParents[l].length + p2] *
                                       this.rawData[i][this.xyParents[l][p1]] * this.rawData[i][this.xyParents[l][p2]];
                }
                dataContrib += this.qS[d][i] * contr;
            }
            if (dataContrib > 0.)
                this.upsilon[d][l] = sumQS / dataContrib;
            else
                this.upsilon[d][l] = Double.MAX_VALUE;
        }
    }

    public Object clone() {
        VariationalSemMixture output = new VariationalSemMixture(this.dataSet,  this.originalGraph,
                                                                 this.freeVariables, this.numMixtures);
        backupParams(output);
        return output;
    }

    protected void backupParams(VariationalSemMixture output) {
        output.score = this.score;
        output.usePriors = this.usePriors; output.singleInvCovParam = this.singleInvCovParam;
        output.betaMuInvCovPriorParam = this.betaMuInvCovPriorParam;
        //output.invPsiA = this.invPsiA; output.invPsiB = this.invPsiB;
        for (int y = 0; y < this.numVars; y++) {
            output.qInvPsiA[y] = this.qInvPsiA[y]; output.qInvPsiB[y] = this.qInvPsiB[y];
            output.qEInvPsi[y] = this.qEInvPsi[y];
        }
        for (int d = 0; d < this.numMixtures; d++) {
            output.alphaMStar[d] = this.alphaMStar[d];
            output.qAlphaM[d] = this.qAlphaM[d];
            output.sumS[d] = this.sumS[d];
            for (int x = 0; x < this.numLatents; x++) {
                for (int p1 = 0; p1 < this.xxParents[x].length; p1++) {
                    output.betaMuMeanPrior[d][x] = this.betaMuMeanPrior[d][x];
                    output.qBetaMuMean[d][x][p1] = this.qBetaMuMean[d][x][p1];
                    output.qEBM[d][x][p1] = this.qEBM[d][x][p1];
                    for (int p2 = 0; p2 < this.xxParents[x].length; p2++) {
                        output.betaMuInvCovPrior[d][x][p1][p2] = this.betaMuInvCovPrior[d][x][p1][p2];
                        output.betaMuCovPrior[d][x][p1][p2] = this.betaMuCovPrior[d][x][p1][p2];
                        output.qBetaMuCov[d][x][p1][p2] = this.qBetaMuCov[d][x][p1][p2];
                        output.qEBMBM[d][x][p1][p2] = this.qEBMBM[d][x][p1][p2];
                    }
                }
                output.upsilon[d][x] = this.upsilon[d][x];
            }
            for (int x1 = 0; x1 < this.numLatents; x1++) {
                output.xMeanPrior[d][x1] = this.xMeanPrior[d][x1];
                for (int x2 = 0; x2 < this.numLatents; x2++)
                    output.xInvCovPrior[d][x1][x2] = this.xInvCovPrior[d][x1][x2];
            }
            for (int y = 0; y < this.numVars; y++)
                 output.lambdaInvCovPriorParam[y] = this.lambdaInvCovPriorParam[y];
            for (int y = 0; y < this.numVars; y++) {
                int numParents = this.yyParents[y].length + this.yxParents[y].length;
                for (int p1 = 0; p1 < numParents; p1++) {
                    output.lambdaMeanPrior[d][y][p1] = this.lambdaMeanPrior[d][y][p1];
                    output.qLambdaMean[d][y][p1] = this.qLambdaMean[d][y][p1];
                    output.qELambda[d][y][p1] = this.qELambda[d][y][p1];
                    for (int p2 = 0; p2 < numParents; p2++) {
                        output.lambdaInvCovPrior[d][y][p1][p2] = this.lambdaInvCovPrior[d][y][p1][p2];
                        output.lambdaCovPrior[d][y][p1][p2] = this.lambdaCovPrior[d][y][p1][p2];
                        output.qLambdaCov[d][y][p1][p2] = this.qLambdaCov[d][y][p1][p2];
                        output.qELambdaLambda[d][y][p1][p2] = this.qELambdaLambda[d][y][p1][p2];
                    }
                }
            }

            for (int i = 0; i < this.sampleSize; i++) {
                output.qS[d][i] = this.qS[d][i];
                for (int x1 = 0; x1 < this.numLatents; x1++) {
                    output.qXMean[d][i][x1] = this.qXMean[d][i][x1];
                    output.qEX[d][i][x1] = this.qEX[d][i][x1];
                    for (int x2 = 0; x2 < this.numLatents; x2++) {
                        output.qXCov[d][i][x1][x2] = this.qXCov[d][i][x1][x2];
                        output.qEXX[d][i][x1][x2] = this.qEXX[d][i][x1][x2];
                    }
                }
                for (int x1 = 0; x1 < this.numMainLatents; x1++)
                    for (int x2 = 0; x2 < this.numMainLatents; x2++)
                        output.qMainXCov[d][i][x1][x2] = this.qMainXCov[d][i][x1][x2];
                for (int y = 0; y < this.numVars; y++)
                    output.likelihoodKernels[d][i][y] = this.likelihoodKernels[d][i][y];
            }
        }
        output.usePriors = this.usePriors;
        output.updatePriors = this.updatePriors;
        output.givenPriors = this.givenPriors;
        output.singleInvCovParam = this.singleInvCovParam;
    }

    public double[][][][] getLatentExpectations() {
        return this.qEXX;
    }

    public double[][][][] getMainLatentExpectations() {
        double output[][][][] = new double[this.numMixtures][this.sampleSize][this.numMainLatents][this.numMainLatents];
        for (int d = 0; d < this.numMixtures; d++)
             for (int i = 0; i < this.sampleSize; i++)
                 for (int x1 = 0; x1 < this.numMainLatents; x1++)
                     for (int x2 = 0; x2 < this.numMainLatents; x2++)
                         output[d][i][x1][x2] = this.qEXX[d][i][x1][x2];
        return output;
    }

    public double[][] getComponentDistribution() {
        return this.qS;
    }

    public double getInvBetaPrior() {
        return this.betaMuInvCovPriorParam;
    }

    public List getLambdaPriors() {
        List output = new ArrayList();
        for (int i = 0; i < this.numVars; i++) {
            output.add(this.indicators[i].toString());
            output.add(new Double(this.lambdaInvCovPriorParam[i]));
        }
        return output;
    }

    public void printPi() {
        double total = 0.;
        for (int i = 0; i < this.numMixtures; i++)
             total += this.qAlphaM[i];
        for (int i = 0; i < this.numMixtures; i++)
            System.out.print(this.qAlphaM[i] / total + " ");
        System.out.println();
    }

    public void latentEdgeDemolition() {
        List<Edge> toRemove = new ArrayList<Edge>();
        for (Iterator it = this.graph.getEdges().iterator(); it.hasNext();) {
            Edge edge = (Edge) it.next();
            if (edge.getNode1().getNodeType() == NodeType.LATENT && edge.getNode2().getNodeType() == NodeType.LATENT)
                toRemove.add(edge);
        }
        this.graph.removeEdges(toRemove);
        for (int i = 0; i < this.numMainLatents; i++) {
            this.xxParents[i] = new int[0];
            for (int d = 0; d < this.numMixtures; d++) {
                this.betaMuMeanPrior[d][i] = new double[0];
                this.betaMuInvCovPrior[d][i] = new double[0][0];
                this.betaMuCovPrior[d][i] = new double[0][0];
            }
        }
        this.score = fEval();
    }

}


